import numpy as np
import sklearn.decomposition as skdecomp
from db import dbfunctions as dbfn
import tables
#import test_reconst_bmi_traj as trbt
import matplotlib.pyplot as plt
# import seaborn
# seaborn.set_context('talk',font_scale=1)
# seaborn.set_style('whitegrid')
import pickle
import scipy
import os
# te = 3138

# if te==3136:
#     hdf = tables.openFile('grom_data/grom20150427_02.hdf')
# elif te ==3138:
#     hdf = tables.openFile('grom_data/grom20150428_02.hdf')
#hdf = tables.openFile('grom_data/grom20151129_19_te3725.hdf')
#dec = pickle.load(open('grom_data/grom20150422_13_PM04221738.pkl'))
#dec = pickle.load(open('grom_data/grom20151129_16_RMLC11291807.pkl'))
#hdf = tables.openFile(os.path.expandvars('$FA_GROM_DATA/grom20151201_03_te3729.hdf'))
#dec = pickle.load(open(os.path.expandvars('$FA_GROM_DATA/grom20151201_01_RMLC12011916.pkl')))

#ReDecoder = trbt.RerunDecoding(hdf, dec, task='point_mass', drives_neurons=3)
#ReDecoder = trbt.RerunDecoding(hdf, dec, task='bmi_multi')#

########################
########################
#### ANALYSIS FCNS #####
########################
########################
# def FA_k_500ms(): 
#     #Function to determine optimal number of factors for first 500 ms of trial
#     #Separates FA model for each target
#     rew_ix = get_trials_per_min(hdf)
#     spk, targ_pos, targ_ix, reach_time = extract_trials(hdf, rew_ix, ms=500)
#     LL = fit_all_targs(spk, targ_ix,iters=10, max_k = 10)
hdf = []

def FA_k_ALLms(hdf):
    #Function to determine optimal number of factors for full trial
    #Separates FA model for each target
    drives_neurons_ix0 = 3
    rew_ix = get_trials_per_min(hdf)
    internal_state = hdf.root.task[:]['internal_decoder_state']
    update_bmi_ix = np.nonzero(np.diff(np.squeeze(internal_state[:, drives_neurons_ix0, 0])))[0]+1

    bin_spk, targ_pos, targ_ix, z, zz = extract_trials_all(hdf, rew_ix, update_bmi_ix=update_bmi_ix)
    LL, ax = fit_all_targs(bin_spk, targ_ix, proc_spks = False, iters=20, max_k = 10, return_ax=True) 
    ax[3,0].set_xlabel('Num Factors')
    ax[3,1].set_xlabel('Num Factors')
    for i in range(4):
        ax[i, 0].set_ylabel('Log Lik')
    return LL, ax

def FA_all_targ_ALLms(hdf, max_k=10, iters=20):
    #Function to determine optimal number of factors for full trial
    drives_neurons_ix0 = 3
    rew_ix = get_trials_per_min(hdf)
    internal_state = hdf.root.task[:]['internal_decoder_state']
    update_bmi_ix = np.nonzero(np.diff(np.squeeze(internal_state[:, drives_neurons_ix0, 0])))[0]+1
    bin_spk, targ_pos, targ_ix, z, zz = extract_trials_all(hdf, rew_ix, update_bmi_ix=update_bmi_ix)
    zscore_X, mu = zscore_spks(bin_spk)
    log_lik, psv, ax = find_k_FA(zscore_X, iters=iters, max_k = max_k, plot=True)
    return log_lik, ax

def SVR_slow_fast_ALLms(hdf, factors=5):
    rew_ix = get_trials_per_min(hdf)
    bin_spk, targ_pos, targ_ix, trial_ix, reach_time = extract_trials_all(hdf, rew_ix)
    fast_vs_slow_trials(bin_spk, targ_ix, reach_time, factors=factors,plot=True, all_trial_tm=True)

def ind_vs_all_targ_SSAlign(factors=5, hdf=hdf):
    rew_ix = get_trials_per_min(hdf)
    bin_spk, targ_pos, targ_ix, trial_ix, reach_time = extract_trials_all(hdf, rew_ix)
    overlap = targ_vs_all_subspace_align(bin_spk, targ_ix, factors=factors)
    f, ax = plt.subplots()
    c = ax.pcolormesh(overlap)
    plt.colorbar(c)
    return overlap


def learning_curve_metrics(hdf_list, epoch_size=56, n_factors=5):
    #hdf_list = [3822, 3834, 3835, 3840]
    #obstacle learning: hdf_list = [4098, 4100, 4102, 4104, 4114, 4116, 4118, 4119]
    rew_ix_list = []
    te_refs = []
    rpm_list = []
    hdf_dict = {}
    perc_succ = []
    time_list = []
    offs = 0

    #f, ax = plt.subplots()
    for te in hdf_list:
        hdf_t = dbfn.TaskEntry(te)
        hdf = hdf_t.hdf
        hdf_dict[te] = hdf

        rew_ix, rpm = pa.get_trials_per_min(hdf, nmin=2,rew_per_min_cutoff=0, 
            ignore_assist=True, return_rpm=True)
        ix = 0
        #ax.plot(rpm)

        trial_ix = np.array([i for i in hdf.root.task_msgs[:] if 
            i['msg'] in ['reward','timeout_penalty','hold_penalty','obstacle_penalty'] ], dtype=hdf.root.task_msgs.dtype)


        while (ix+epoch_size) < len(rew_ix):
            start_rew_ix = rew_ix[ix]
            end_rew_ix = rew_ix[ix+epoch_size]
            msg_ix_mod = np.nonzero(scipy.logical_and(trial_ix['time']<=end_rew_ix, trial_ix['time']>start_rew_ix))[0]
            all_msg = trial_ix[msg_ix_mod]
            perc_succ.append(len(np.nonzero(all_msg['msg']=='reward')[0]) / float(len(all_msg)))

            rew_ix_list.append(rew_ix[ix:ix+epoch_size])
            rpm_list.append(np.mean(rpm[ix:ix+epoch_size]))
            te_refs.append(te)
            time_list.append((0.5*(start_rew_ix+end_rew_ix))+offs)

            ix += epoch_size
        offs = offs+len(hdf.root.task)

    #For each epoch, fit FA model (stick w/ 5 factors for now):
    ratio = []
    for te, r_ix in zip(te_refs, rew_ix_list):
        print(te, len(r_ix))

        update_bmi_ix = np.nonzero(np.diff(np.squeeze(hdf.root.task[:]['internal_decoder_state'][:, 3, 0])))[0] + 1
        bin_spk, targ_pos, targ_ix, z, zz = pa.extract_trials_all(hdf_dict[te], r_ix, time_cutoff=1000, update_bmi_ix=update_bmi_ix)
        zscore_X, mu = pa.zscore_spks(bin_spk)
        FA = skdecomp.FactorAnalysis(n_components=n_factors)
        FA.fit(zscore_X)

        #SOT Variance Ratio by target
        #Priv var / mean
        Cov_Priv = np.sum(FA.noise_variance_)
        U = np.mat(FA.components_).T
        Cov_Shar = np.trace(U*U.T)

        ratio.append(Cov_Shar/(Cov_Shar+Cov_Priv))


########################
########################
###### HELPER FCNS #####
########################
########################

#Get behavior
def get_trials_per_min(hdf,nmin=2, rew_per_min_cutoff=0, ignore_assist=False, return_rpm=False, return_per_succ=False):
    '''
    Summary: Getting trials per minute from hdf file
    Input param: hdf: hdf file to use
    Input param: nmin: number of min to use a rectangular window
    Input param: rew_per_min_cutoff: ignore rew_ix after a 
        certain rew_per_min low threshold is passed
    Output param: rew_ix = rewarded indices in hdf file
    '''

    rew_ix = np.array([t[1] for it, t in enumerate(hdf.root.task_msgs[:]) if t[0]=='reward'])
    tm = np.zeros((np.max(rew_ix)+1))
    tm[rew_ix] += 1
    
    if hasattr(hdf.root.task, 'assist_level'):
        assist_ix = np.nonzero(hdf.root.task[:]['assist_level']==0)[0]
    else:
        assist_ix = np.zeros((len(hdf.root.task)))

    #Each row occurs ~1/60 sec, so: 
    minute = 60*60;
    min_wind = np.ones((nmin*minute))/float(nmin)
    rew_per_min_tmp = np.convolve(min_wind, tm, mode='same')

    #Now smooth please: 
    smooth_wind = np.ones((3*minute))/float(3*minute)
    rew_per_min = pk_convolve(smooth_wind, rew_per_min_tmp)

    if rew_per_min_cutoff > 0:
        ix = np.nonzero(rew_per_min < rew_per_min_cutoff)[0]
        if len(ix)>0:
            cutoff_ix = ix[0]
        else:
            cutoff_ix = rew_ix[-1]
    
    else:
        cutoff_ix = rew_ix[-1]

    if ignore_assist:
        try:
            beg_zer_assist_ix = assist_ix[0]
        except:
            print('No values w/o assist for filename: ', hdf.filename)
            beg_zer_assist_ix = rew_ix[-1]+1
    else:
        beg_zer_assist_ix = 0


    #plt.plot(np.arange(len(tm))/float(minute), rew_per_min)
    ix_final = scipy.logical_and(rew_ix <= cutoff_ix, rew_ix >= beg_zer_assist_ix)
    if return_rpm:
        return rew_ix[ix_final], rew_per_min[rew_ix[ix_final]]
    else:
        return rew_ix[ix_final]

def pk_convolve(window, arr):
    '''
    Summary: Same as NP convolve but no edge effects (edges padded with zeros+start value and zeros+end value )
    Input param: window: window to convolve
    Input param: arr: array (longer than window)
    Output param: 
    '''
    win_len = len(window)
    start = arr[0]
    end = arr[-1]

    arr_pad = np.hstack(( np.zeros((win_len, )) + start, arr, np.zeros((win_len, )) + end ))
    tmp = np.convolve(window, arr_pad, mode='same')
    return tmp[win_len:-win_len]


def extract_trials_all(hdf, rew_ix, neural_bins = 100, time_cutoff=40, hdf_ix=False, 
    update_bmi_ix=None, rew_pls=False, step_dict=None):
    '''
    Summary: method to extract all time points from trials
    Input param: hdf: task file input
    Input param: rew_ix: rows in the hdf file corresponding to reward times
    Input param: neural_bins: ms per bin
    Input param: time_cutoff: time in minutes, only extract trials before this time
    Input param: hdf_ix: bool, whether to return hdf row corresponding to time of decoder 
    update (and hence end of spike bin)
    Input param: update_bmi_ix: bins for updating bmi

    Input param: rew_plx : If rew_ix is actually a N x 2 array with error trials included (default = false)
    input param: step_dict: Teh number of steps to go backward from each type of trial in rew_ix (only 
        if rew_pls is True)

    Output param: bin_spk -- binned spikes in time x units
                  targ_i_all -- target location at each update
                  targ_ix -- target index 
                  trial_ix -- trial number
                  reach_time -- reach time for trial
                  hdf_ix -- end bin in units of hdf rows
    '''
    if update_bmi_ix is None:
        update_bmi_ix = ReDecoder.update_bmi_ix

    it_cutoff = time_cutoff*60*60

    #Get Go cue and 

    if rew_pls:
        go_ix = np.array([hdf.root.task_msgs[it - step_dict[m['msg']]]['time'] for it, m in enumerate(hdf.root.task_msgs[:]) 
            if m['msg'] in list(step_dict.keys())])

        rew_ix = np.array([hdf.root.task_msgs[it]['time'] for it, m in enumerate(hdf.root.task_msgs[:]) 
            if m['msg'] in list(step_dict.keys())])

        outcome_ix = np.array([hdf.root.task_msgs[it]['msg'] for it, m in enumerate(hdf.root.task_msgs[:]) 
            if m['msg'] in list(step_dict.keys())])

    else:
        go_ix = np.array([hdf.root.task_msgs[it-3][1] for it, t in enumerate(hdf.root.task_msgs[:]) if 
            scipy.logical_and(t[0] == 'reward', t[1] in rew_ix)])
    

    go_ix = go_ix[go_ix<it_cutoff]
    rew_ix = rew_ix[go_ix < it_cutoff]

    targ_i_all = np.array([[-1, -1]])
    trial_ix_all = np.array([-1])
    reach_tm_all = np.array([-1])
    hdf_ix_all = np.array([-1])

    bin_spk = np.zeros((1, hdf.root.task[0]['spike_counts'].shape[0]))-1

    def bin_spks(spk_i, g_ix, r_ix):
        #Need to use 'update_bmi_ix' from ReDecoder to get bin edges correctly:
        trial_inds = np.arange(g_ix, r_ix+1)
        end_bin = np.array([(j,i) for j, i in enumerate(trial_inds) if np.logical_and(i in update_bmi_ix, i>=(g_ix+5))])
        nbins = len(end_bin)
        bin_spk_i = np.zeros((nbins, spk_i.shape[1]))

        hdf_ix_i = []
        for ib, (i_ix, hdf_ix) in enumerate(end_bin):
            #Inclusive of EndBin
            bin_spk_i[ib,:] = np.sum(spk_i[i_ix-5:i_ix+1,:], axis=0)
            hdf_ix_i.append(hdf_ix)
        return bin_spk_i, nbins, np.array(hdf_ix_i)

    for ig, (g, r) in enumerate(zip(go_ix, rew_ix)):
        spk_i = hdf.root.task[g:r]['spike_counts'][:,:,0]

        #Sum spikes in neural_bins:
        bin_spk_i, nbins, hdf_ix_i = bin_spks(spk_i, g, r)
        bin_spk = np.vstack((bin_spk, bin_spk_i))
        targ_i_all = np.vstack(( targ_i_all, np.tile(hdf.root.task[g+1]['target'][[0,2]], (bin_spk_i.shape[0], 1)) ))
        trial_ix_all = np.hstack(( trial_ix_all, np.zeros(( bin_spk_i.shape[0] ))+ig ))
        reach_tm_all = np.hstack((reach_tm_all, np.zeros(( bin_spk_i.shape[0] ))+((r-g)*1000./60.) ))
        hdf_ix_all = np.hstack((hdf_ix_all, hdf_ix_i ))

    print(go_ix.shape, rew_ix.shape, bin_spk.shape, bin_spk_i.shape, nbins, hdf_ix_i.shape)
    targ_ix = get_target_ix(targ_i_all[1:,:])
    
    if hdf_ix:
        return bin_spk[1:,:], targ_i_all[1:,:], targ_ix, trial_ix_all[1:], reach_tm_all[1:], hdf_ix_all[1:]
    else:
        return bin_spk[1:,:], targ_i_all[1:,:], targ_ix, trial_ix_all[1:], reach_tm_all[1:]

def extract_trials(hdf, rew_ix, ms=500, time_cutoff=40):
    it_cutoff = time_cutoff*60*60
    nsteps = int(ms/(1000.)*60)

    #Get Go cue and 
    go_ix = np.array([hdf.root.task_msgs[it-3][1] for it, t in enumerate(hdf.root.task_msgs[:]) if t[0] == 'reward'])
    go_ix = go_ix[go_ix<it_cutoff]

    spk = np.zeros((len(go_ix), nsteps, hdf.root.task[0]['spike_counts'].shape[0]))

    targ_pos = np.zeros((len(go_ix), 2))
    reach_time = np.zeros((len(go_ix), ))
    for ig, g in enumerate(go_ix):
        spk[ig, :, :] = hdf.root.task[g:g+nsteps]['spike_counts'][:,:,0]
        targ_pos[ig, :] = hdf.root.task[g+1]['target'][[0,2]]
        reach_time[ig] = (rew_ix[ig]- g)/(60.) #In seconds
    
    targ_ix = get_target_ix(targ_pos)
    return spk, targ_pos, targ_ix, reach_time

def get_target_ix(targ_pos):
    ''' Input: targ_pos - n x 2 array
        Output: n x 1 array of indices
    '''
    #Target Index: 
    b = np.ascontiguousarray(targ_pos).view(np.dtype((np.void, targ_pos.dtype.itemsize * targ_pos.shape[1])))
    _, idx = np.unique(b, return_index=True)
    unique_targ = targ_pos[idx,:]

    #Order by theta: 
    theta = np.arctan2(unique_targ[:,1],unique_targ[:,0])
    thet_i = np.argsort(theta)
    unique_targ = unique_targ[thet_i, :]
    
    targ_ix = np.zeros((targ_pos.shape[0]), )
    for ig, (x,y) in enumerate(targ_pos):
        targ_ix[ig] = np.nonzero(np.sum(targ_pos[ig,:]==unique_targ, axis=1)==2)[0]
    return targ_ix

def proc_spks(spk, targ_ix, targ_ix_analysis=0, neural_bins = 100, return_unshapedX=False):
    '''
    Summary: processes bin_spikes (time x units)
    Input param: spk: time x units (in binned spikes)
    Input param: targ_ix: array lenth of bins
    Input param: targ_ix_analysis: 
    Input param: neural_bins:
    Input param: return_unshapedX:
    Output param: 
    '''

    if targ_ix_analysis == 'all':
        spk_trunc = spk.copy()
    else:
        ix = np.nonzero(targ_ix==targ_ix_analysis)[0]
        spk_trunc = spk[ix, :, :]

    bin_ix = int(neural_bins/1000.*60.)
    spk_trunc_bin = np.zeros((spk_trunc.shape[0], spk_trunc.shape[1]/bin_ix, spk_trunc.shape[2]))
    for i_start in range(spk_trunc.shape[1]/bin_ix):
        spk_trunc_bin[:, i_start, :] = np.sum(spk_trunc[:, (i_start*bin_ix):(i_start+1)*bin_ix, :], axis=1)
    resh_spk_trunc_bin = spk_trunc_bin.reshape((spk_trunc_bin.shape[0]*spk_trunc_bin.shape[1], spk_trunc_bin.shape[2]))
    if return_unshapedX:
        return spk_trunc_bin
    else:
        return resh_spk_trunc_bin

def zscore_spks(proc_spks):
    '''
    proc_spks in time x neurons
    zscore_X in time x neurons
    '''

    mu = np.tile(np.mean(proc_spks, axis=0), (proc_spks.shape[0], 1))
    zscore_X = proc_spks - mu
    return zscore_X, mu

def find_k_FA(zscore_X, xval_test_perc = .1, iters=100, max_k = 20, plot=True):
    ntrials = zscore_X.shape[0]
    ntrain = ntrials*(1-xval_test_perc)
    log_lik = np.zeros((iters, max_k))
    perc_shar_var = np.zeros((iters, max_k))

    for i in range(iters):
        print('iter: ', i)
        ix = np.random.permutation(ntrials)
        train_ix = ix[:ntrain]
        test_ix = ix[ntrain:]

        for k in range(max_k):
            print('factor n : ', k)
            FA = skdecomp.FactorAnalysis(n_components=k+1)
            FA.fit(zscore_X[train_ix,:])
            LL = np.sum(FA.score(zscore_X[test_ix,:]))
            log_lik[i, k] = LL

            if (k+1) == max_k:
                cov_shar = np.matrix(FA.components_) * np.matrix(FA.components_.T)
                perc_shar_var[i, :] = np.cumsum(np.diag(cov_shar)) 

    log_lik[log_lik<-1e4] = np.nan
    if plot:
        fig, ax = plt.subplots(nrows=2)
        mu = 1/float(log_lik.shape[0])*np.nansum(log_lik,axis=0)
        
        for it in range(log_lik.shape[0]):
            for nf in range(log_lik.shape[1]):
                if np.isnan(log_lik[it, nf]):
                    log_lik[it, nf] = mu[nf]

        fin = np.tile(np.array([perc_shar_var[:,-1]]).T, [1, perc_shar_var.shape[1]])
        perc = perc_shar_var/fin.astype(float)

        ax[0].errorbar(np.arange(1,log_lik.shape[1]+1), np.mean(log_lik,axis=0), yerr=np.std(log_lik,axis=0)/np.sqrt(iters), fmt='o')
        #plt.plot(np.arange(1,log_lik.shape[1]+1), np.mean(log_lik,axis=0),'.-')
        ax[0].set_ylabel('Log Likelihood of Held Out Data')

        ax[1].errorbar(np.arange(1,perc.shape[1]+1), np.mean(perc,axis=0), yerr=np.std(perc,axis=0)/np.sqrt(iters), fmt='o')
        ax[1].set_xlabel('Number of Factors')
        ax[1].set_ylabel('Perc. Shar. Var. ')
        return log_lik, perc_shar_var, ax
    else:

        return log_lik, perc_shar_var

def fit_all_targs(spk, targ_ix, proc_spks=True, iters=10, max_k = 10, return_ax=False):
    LL = dict()
    tg = np.unique(targ_ix)
    fig, ax = plt.subplots(nrows=4, ncols=2)

    for it, t in enumerate(tg):
        if proc_spks:
            resh_spk_trunc_bin = proc_spks(spk, targ_ix, targ_ix_analysis=t)
        else:
            ix = np.nonzero(targ_ix==t)[0]
            resh_spk_trunc_bin = spk[ix,:]
            print('ix: ', str(len(ix)))
        zscore_X, mu = zscore_spks(resh_spk_trunc_bin)
        log_lik, psv = find_k_FA(zscore_X, iters=iters, max_k =max_k, plot=False)
        LL[t] = log_lik
        #TODO: insert way to 
        ax[it%4, it/4].plot(np.arange(10)+1, np.mean(log_lik, axis=0), '.-', label='Targ '+str(t))
        ax[it%4, it/4].errorbar(np.arange(1,log_lik.shape[1]+1), np.mean(log_lik,axis=0), yerr=np.std(log_lik,axis=0)/np.sqrt(iters), fmt='o')
        ax[it%4, it/4].set_title('Targ '+str(t))
    
    plt.axis('tight')
    plt.tight_layout()
    if return_ax:
        return LL, ax
    else:
        return LL


def shared_vs_total_var(spk, targ_ix, reach_time, factors=5):
    #For each target, fit a FA analysis model:
    tg = np.unique(targ_ix)
    ratio = np.zeros((len(tg), ))
    beh = ratio.copy()
    for it, t in enumerate(tg):
        zscore_X, z, m = proc_spks(spk, targ_ix, targ_ix_analysis=t)
        FA = skdecomp.FactorAnalysis(n_components=factors)
        FA.fit(zscore_X)

        Cov_Priv = np.sum(FA.noise_variance_)
        U = np.mat(FA.components_).T
        Cov_Shar = np.trace(U*U.T)

        ratio[it] = Cov_Shar/(Cov_Shar+Cov_Priv)

        ix = np.nonzero(targ_ix==t)[0]
        beh[it] = np.mean(reach_time[ix])

    f, ax1 = plt.subplots()
    ax1.plot(ratio, label='Ratio')
    ax1.set_ylabel('Shared to Total Variance')
    ax2 = ax1.twinx()
    ax2.plot(beh, 'k-', label='Reach Time (sec)')
    ax2.set_ylabel('Reach Time')
    ax2.set_title('N Factors: '+str(factors))
    ax1.set_xlabel('Target Number')
    plt.legend()
    return ratio, beh

def fast_vs_slow_trials(spk, targ_ix, reach_time, factors=5,plot=True, all_trial_tm=False):
    tg = np.unique(targ_ix)

    fast_ratio = np.zeros((len(tg), ))
    slow_ratio = fast_ratio.copy()
    beh_fast = fast_ratio.copy()
    beh_slow = fast_ratio.copy()

    for it, t in enumerate(tg):
        ix = np.nonzero(targ_ix==t)[0]
        mid = np.percentile(reach_time[ix], 50)
        fast_ix_ix = np.nonzero(reach_time[ix]<= mid)[0]
        slow_ix_ix = np.nonzero(reach_time[ix]> mid)[0]

        #Fast FA: 
        def get_ratio(ix_ix):
            if all_trial_tm:
                X = spk[ix[ix_ix], :]
                zscore_X, mu = zscore_spks(X)
            else:
                X = proc_spks(spk, targ_ix, targ_ix_analysis=t,return_unshapedX=True)
                X = X[ix_ix, :, :]
                zscore_X = zscore_spks(X.reshape((X.shape[0]*X.shape[1], X.shape[2])))
                
            FA = skdecomp.FactorAnalysis(n_components=factors)
            FA.fit(zscore_X)

            Cov_Priv = np.sum(FA.noise_variance_)
            U = np.mat(FA.components_).T
            Cov_Shar = np.trace(U*U.T)
            return Cov_Shar/(Cov_Shar+Cov_Priv)
        
        fast_ratio[it] = get_ratio(fast_ix_ix)
        slow_ratio[it] = get_ratio(slow_ix_ix)
        beh_fast[it] = np.mean(reach_time[ix][fast_ix_ix])
        beh_slow[it] = np.mean(reach_time[ix][slow_ix_ix])

    if plot:
        f, ax1 = plt.subplots(nrows=4, ncols=2)
        for it in range(len(tg)):
            ii = it%4
            jj = it/4

            ax1[ii, jj].plot(tg[it]+np.array([-.33, .33]), [fast_ratio[it], slow_ratio[it]], 'r*-')
            #ax1[ii, jj].plot(tg[it]+.33, slow_ratio[it], 'b*')
            ax1[ii, jj].set_ylabel('Var. Ratio')
            #ax1[ii, jj].set_ylim([.35, .55])
            ax2 = ax1[ii, jj].twinx()
            ax2.plot(tg[it]+np.array([-.33, .33]), [beh_fast[it], beh_slow[it]], 'bo-', label='Reach Time (sec)')
            #ax2.plot(tg[it]+.33, beh_slow[it], 'bo', label='Slow Reach Time (sec)')

            ax2.set_ylabel('Reach Time')
            ax2.set_title('N Factors: '+str(factors))
            #ax2.set_ylim([2.5, 4.5])
            ax1[ii, jj].set_xlabel('Target Number '+str(tg[it]))
            #plt.legend()
    plt.tight_layout()
    return fast_ratio, slow_ratio, beh_fast, beh_slow

def FA_subspace_align(FA1, FA2):
    U_A = np.mat(FA1.components_.T)
    U_B = np.mat(FA2.components_).T
    v, s, vt = np.linalg.svd(U_B*U_B.T)
    P_B = v*vt
    S_A_shared = U_A*U_A.T
    return np.trace(P_B*S_A_shared*P_B.T)/np.trace(S_A_shared)

def targ_vs_all_subspace_align(bin_spk, targ_ix, factors=5):
    X_zsc, mu = zscore_spks(bin_spk)
    FA_full = skdecomp.FactorAnalysis(n_components = factors)
    FA_full.fit(X_zsc)

    unique_targ = np.unique(targ_ix)
    Overlap = np.zeros((len(unique_targ)+1, len(unique_targ)+1))
    
    FA_targ = dict()
    for it, t in enumerate(unique_targ):
        FA = skdecomp.FactorAnalysis(n_components = factors)
        ix = np.nonzero(targ_ix==t)[0]
        X_zsc, mu = zscore_spks(bin_spk[ix,:])
        FA.fit(X_zsc)
        FA_targ[t] = FA
        Overlap[it, len(unique_targ)] = FA_subspace_align(FA, FA_full)
        Overlap[len(unique_targ), it] = FA_subspace_align(FA_full, FA)

    Overlap[len(unique_targ), len(unique_targ)] = FA_subspace_align(FA_full, FA_full)
        
    for it, t in enumerate(unique_targ):
        for iitt, tt in enumerate(unique_targ):
            Overlap[it, iitt] = FA_subspace_align(FA_targ[t], FA_targ[tt])

    return Overlap








from ..bmi import extractor
import numpy as np

from ..bmi import state_space_models
import datetime
import matplotlib.pyplot as plt

class State(object):
    '''For compatibility with other BMI decoding implementations, literally just holds the state'''

    def __init__(self, mean, *args, **kwargs):
        self.mean = mean

class RatFilter(object):
    '''Moving Avergae Filter used in 1D or 2D LFP control:
    x_{t} = a0*x_{t} + a1*x_{t-1} + a2*x_{t-2} + ...
    x_{t} = b_1:t*x_{}


    Parameters

    ----------
    A: np.array of shape (N, )
        Weights for previous states
    X: np. array of previous states (N, )
    '''

    def __init__(self, task_params):
        self.e1_inds = task_params['e1_inds']
        self.e2_inds = task_params['e2_inds']
        self.FR_to_freq_fn = task_params['FR_to_freq_fn']
        self.t1 = task_params['t1']
        self.t2 = task_params['t2']
        self.mid = task_params['mid']
        self.dec_params = task_params

        #Cursor data (X)
        self.X = 0.

        #Freq data(F) 
        self.F = 0.
        self.baseline = False

    def get_mean(self):
        return np.array(self.state.mean).ravel()

    def init_from_task(self, n_units, **kwargs):
        #Define n_steps
        if 'nsteps' in kwargs:
            self.n_steps = kwargs['nsteps']
            self.A = np.ones(( self.n_steps, ))/float(self.n_steps)

            #Neural data (Y)
            self.Y = np.zeros(( self.n_steps, n_units))
            self.n_units = n_units

        else:
            raise Exception


    def _init_state(self, init_state=None,**kwargs):
        if init_state is None:
            init_state = 0

        self.state = State(init_state)

    def __call__(self, obs, **kwargs):
        self.state = self._mov_avg(obs, **kwargs)

    def _mov_avg(self, obs,**kwargs):
        ''' Function to compute moving average with old mean and new observation'''

        self.Y[:-1, :] = self.Y[1:, :]
        self.Y[-1, :] = np.squeeze(obs)

        d_fr = np.sum(self.Y[:, self.e1_inds], axis=1) - np.sum(self.Y[:, self.e2_inds], axis=1)
        mean_FR = np.dot(d_fr, self.A)
        self.X = mean_FR
        self.F = self.FR_to_freq_fn(self.X)
        return State(self.X)

    def FR_to_freq(self, mean_FR):
        return self.FR_to_freq_fn(mean_FR)

    def _pickle_init(self):
        pass

class IsmoreSleepFilter(RatFilter):
    def __init__(self, task_params):
        self.e1_inds = task_params['e1_inds']
        self.e2_inds = task_params['e2_inds']
        self.FR_to_alpha_fn = task_params['FR_to_alpha_fn']
        self.dec_params = task_params
        self.mid = task_params['mid']
        self.e1_max = task_params['e1_perc']
        self.e2_max = task_params['e2_perc']
        self.freq_lim = task_params['freq_lim']

        #Cursor data (X)
        self.FR = 0.

        #Freq data(F) 
        self.alpha = 0.
        self.model_attrs = []
        self.baseline = False

    def init_from_task(self, **kwargs):
        #Define n_steps
        self.n_steps = kwargs.pop('nsteps', 1)
        self.A = np.ones(( self.n_steps, ))/float(self.n_steps)

        #Neural data (Y)
        self.Y = np.zeros(( self.n_steps, len(self.e1_inds)+len(self.e2_inds)))
        self.n_units = len(self.e1_inds)+len(self.e2_inds)

    def _mov_avg(self, obs,**kwargs):
        ''' Function to compute moving average with old mean and new observation'''

        self.Y[:-1, :] = self.Y[1:, :]
        self.Y[-1, :] = np.squeeze(obs)

        if self.e1_max is not None:
            e1_tmp = np.min([self.e1_max, np.sum(self.Y[:, self.e1_inds], axis=1)])
        else:
            e1_tmp = np.sum(self.Y[:, self.e1_inds], axis=1)

        if self.e2_max is not None:
            e2_tmp = np.min([self.e2_max, np.sum(self.Y[:, self.e2_inds], axis=1)])
        else:
            e2_tmp = np.sum(self.Y[:, self.e2_inds], axis=1)

        d_fr = e1_tmp - e2_tmp
        mean_FR = np.dot(d_fr, self.A)
        self.FR = mean_FR
        self.alpha = self.FR_to_alpha_fn(self.FR)

        # Max alpha is -1 or 1 : 
        if self.alpha > self.freq_lim[1]:
            self.alpha = self.freq_lim[1]
        elif self.alpha < self.freq_lim[0]:
            self.alpha = self.freq_lim[0]

        self.baseline = self.FR < self.mid
        return State(self.alpha)

from riglib.bmi.bmi import Decoder
class RatDecoder(Decoder):

    def __init__(self, *args, **kwargs):
        
        #Args: filter, units, ssm, extractor_cls, extractor_kwargs
        super(RatDecoder, self).__init__(args[0], args[1], args[2])
        self.extractor_cls = args[3]
        self.extractor_kwargs = args[4]
        self.n_features = len(self.filt.e1_inds) + len(self.filt.e2_inds)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self,key,value)

    def predict(self, neural_obs, **kwargs):
        self.filt(neural_obs, **kwargs)


    def init_from_task(self,**kwargs):
        pass

class IsmoreSleepDecoder(RatDecoder):
    def __init__(self, *args, **kwargs):
        self.binlen = 0.1
        super(IsmoreSleepDecoder, self).__init__(*args, **kwargs)

########## Functions to make decoder ###########

def create_decoder(ssm, task_params):
    filter_ = RatFilter(task_params)
    decoder = RatDecoder(filter_, task_params['units'], ssm, task_params['extractor_cls'], dict())
    return decoder

########### Called from trainbmi.py to make decoder from Baseline #####
import re
cellname = re.compile(r'(\d{1,3})\s*(\w{1})')

def calc_decoder_from_baseline_file(neural_features, neural_features_unbinned, units, nsteps, prob_t1,
    prob_t2, timeout, timeout_pause, freq_lim, e1_inds, e2_inds, sim_fcn='rat', **kwargs):

    #Enter e1, e2 as string: 
    if np.logical_or(e1_inds is None, e2_inds is None):
        e1_string = input('Enter e1 cells: ')
        e2_string = input('Enter e2 cells: ')
        
        e1 = np.array([ (int(c), ord(u) - 96) for c, u in cellname.findall(e1_string)])
        e2 = np.array([ (int(c), ord(u) - 96) for c, u in cellname.findall(e2_string)])

        e1_inds = np.array([i for i, u in enumerate(units) if np.logical_and(u[0] in e1[:,0], u[1] in e1[:,1])])
        e2_inds = np.array([i for i, u in enumerate(units) if np.logical_and(u[0] in e2[:,0], u[1] in e2[:,1])])

    T = neural_features.shape[0]
    if 'saturate_perc' in kwargs:
        baseline_data = np.zeros((T - nsteps, 2))
    else:
        baseline_data = np.zeros((T - nsteps))
    for ib in range(T-nsteps):
        if 'saturate_perc' in kwargs:
            baseline_data[ib, 0] = np.mean(np.sum(neural_features[ib:ib+nsteps, 
                e1_inds], axis=1))
            baseline_data[ib, 1] = np.mean(np.sum(neural_features[ib:ib+nsteps, 
                e2_inds], axis=1)) 
        else:
            baseline_data[ib] = np.mean(np.sum(neural_features[ib:ib+nsteps, 
                e1_inds], axis=1))-np.mean(np.sum(neural_features[ib:ib+nsteps, 
                e2_inds], axis=1))

    if 'saturate_perc' in kwargs:

        sat_perc = kwargs.pop('saturate_perc')
        # ignore the first second of data
        e1_perc = np.percentile(baseline_data[20:, 0], sat_perc)
        e2_perc = np.percentile(baseline_data[20:, 1], sat_perc)

        baseline_data[:, 0][baseline_data[:, 0] > e1_perc] = e1_perc
        baseline_data[:, 1][baseline_data[:, 1] > e2_perc] = e2_perc
        baseline_data = baseline_data[:, 0] - baseline_data[:, 1]
    else:
        e1_perc = None
        e2_perc = None

    x, pdf, pdf_individual = generate_gmm(baseline_data)

    if sim_fcn == 'rat':
        t2, mid, t1, num_t1, num_t2, num_miss, FR_to_freq_fn = sim_data(x, pdf, pdf_individual, prob_t1, prob_t2,
        baseline_data, timeout, timeout_pause, freq_lim, sim_bmi_fcn='rat')

        return e1_inds, e2_inds, FR_to_freq_fn, units, t1, t2, mid
    
    elif sim_fcn == 'ismore':
        # Get fcn: 
        t1 = prob_under_pdf(x, pdf, prob_t1)
        t2 = prob_under_pdf(x, pdf, prob_t2)
        idx_mid = np.argmax(pdf)
        mid = x[idx_mid]
        FR_to_alpha_fn = map_to_freq(t2, mid, t1, freq_lim[0], freq_lim[1])
        
        # Plot FR to alpha fcn
        x_axis = np.linspace(t2, t1, 100)
        y_axis = []
        for xi in x_axis:
            yi = FR_to_alpha_fn(xi)
            yi = np.max([freq_lim[0], yi])
            yi = np.min([freq_lim[1], yi])
            y_axis.append(yi)

        x_axis2 = np.arange(-10, 10)
        y_axis2 = []
        for xi in x_axis2:
            y_axis2.append(FR_to_alpha_fn(xi))

        import matplotlib.pyplot as plt
        f, ax = plt.subplots()
        ax.plot(x_axis, y_axis)
        ax.plot(x_axis2, y_axis2, 'r.')
        ax.plot([-5, 5], [0, 0], 'r--')

        kwargs2 = dict(replay_neural_features = neural_features, e1_inds=e1_inds, 
            e2_inds = e2_inds, FR_to_alpha_fn=FR_to_alpha_fn, mid = mid, e1_perc=e1_perc,
            e2_perc=e2_perc, freq_lim=freq_lim)

        filt = IsmoreSleepFilter(kwargs2)
        decoder = IsmoreSleepDecoder(filt, units, state_space_models.StateSpaceEndptPos1D(),
            extractor.BinnedSpikeCountsExtractor, {})
        if kwargs['skip_sim']:
            nrewards = []
            import pdb
            pdb.set_trace()

        else:
            if type(kwargs['targets_matrix']) is str:
                import pickle
                kwargs['targets_matrix'] = pickle.load(open(kwargs['targets_matrix']))
            pname = ismore_sim_bmi(neural_features_unbinned, decoder, targets_matrix=kwargs['targets_matrix'],
                session_length=kwargs['session_length'])

            # Analyze data: 
            import tables
            import matplotlib.pyplot as plt
            hdf = tables.openFile(pname[:-4]+'.hdf')

            # Plot x/y trajectory: 
            f, ax = plt.subplots()
            ax.plot(hdf.root.task[2:]['plant_pos'][:, 0], hdf.root.task[2:]['plant_pos'][:, 1])

            ix = np.nonzero(hdf.root.task[:]['target_index'] == 1)[0]
            ax.plot(hdf.root.task[ix[0]]['target_pos'][0], hdf.root.task[ix[0]]['target_pos'][1], 'r.',
                markersize=20)
        
            ix = np.nonzero(hdf.root.task[:]['target_index'] == 0)[0] + 5
            ax.plot(hdf.root.task[ix[0]]['target_pos'][0], hdf.root.task[ix[0]]['target_pos'][1], 'g.',
                markersize=20)

            nrewards = np.nonzero(hdf.root.task_msgs[:]['msg']=='reward')[0]

        return decoder, len(nrewards)


###### From Rat BMI #######
###### From Rat BMI #######
###### From Rat BMI #######
###### From Rat BMI #######

from sklearn import metrics
from sklearn.mixture import GMM
import numpy as np
import matplotlib.pyplot as plt

def generate_gmm(data, ax=None):
    ##reshape the data
    X = data.reshape(data.shape[0], 1)
    ##fit models with 1-10 components
    N = np.arange(1,11)
    models = [None for i in range(len(N))]
    for i in range(len(N)):
        models[i] = GMM(N[i]).fit(X)
    ##compute AIC
    AIC = [m.aic(X) for m in models]
    ##figure out the best-fit mixture
    M_best = models[np.argmin(AIC)]
    x = np.linspace(data.min()-1, data.max()+1, data.size)
    ##compute the pdf
    logprob, responsibilities = M_best.score_samples(x.reshape(x.size, 1))
    pdf = np.exp(logprob)
    pdf_individual = responsibilities * pdf[:, np.newaxis]
    #plot the stuff
    if ax is None:
        fig, ax = plt.subplots()
    ax.hist(X, 50, normed = True, histtype = 'stepfilled', alpha = 0.4)
    ax.plot(x, pdf, '-k')
    ax.plot(x, pdf_individual, '--k')
    ax.text(0.04, 0.96, "Best-fit Mixture",
        ha='left', va='top', transform=ax.transAxes)
    ax.set_xlabel('$x$')
    ax.set_ylabel('$p(x)$')
    return x, pdf, pdf_individual

##this function takes in an array of x-values and an array
##of y-values that correspond to a probability density function
##and determines the x-value at which the area under the PDF is approximately
##equal to some value passed in the arguments.
def prob_under_pdf(x_pdf, y_pdf, prob):
    auc = 0
    i = 2
    while auc < prob:
        x_range = x_pdf[0:i]
        y_range = y_pdf[0:i]
        auc = metrics.auc(x_range, y_range)
        i+=1
    return x_pdf[i]

##function to map ensemble values to frequency values
def map_to_freq(t2, mid, t1, min_freq, max_freq):
    fr_pts = np.array([t2, mid, t1])
    freq_pts = np.array([min_freq, (float(max_freq) + float(min_freq))/2., max_freq])
    z = np.polyfit(fr_pts, freq_pts, 2)
    p = np.poly1d(z)
    return p

def sim_data(x, pdf, pdf_individual, prob_t1, prob_t2, data, 
    timeout, timeout_pause, freq_lim, sim_bmi_fcn='rat'):
    t1 = prob_under_pdf(x, pdf, prob_t1)
    t2 = prob_under_pdf(x, pdf, prob_t2)
    idx_mid = np.argmax(pdf)
    mid = x[idx_mid]
    fig, ax1 = plt.subplots()
    ax1.hist(data+np.random.normal(0, 0.1*data.std(), data.size), 50, 
        normed = True, histtype = 'stepfilled', alpha = 0.4)
    ax1.plot(x, pdf, '-k')
    ax1.plot(x, pdf_individual, '--k')
    ax1.text(0.04, 0.96, "Best-fit Mixture",
        ha='left', va='top', transform=ax1.transAxes)
    ax1.set_xlabel('Cursor Value (E1-E2)')
    ax1.set_ylabel('$p(x)$')
    ##find the points where t1 and t2 lie on the gaussian
    idx_t2 = np.where(x>t2)[0][0]
    x_t2 = t2
    y_t2 = pdf[idx_t2]
    idx_t1 = np.where(x>t1)[0][0]
    x_t1 = t1
    y_t1 = pdf[idx_t1]
    y_mid = pdf[idx_mid]
    ax1.plot(x_t1, y_t1, 'o', color = 'g')
    ax1.plot(x_t2, y_t2, 'o', color = 'g')
    ax1.plot(mid, y_mid, 'o', color = 'g')
    ax1.set_title("Firing rate histogram and gaussian fit")
    ax1.annotate('T1: ('+str(round(x_t1, 3))+')', xy=(x_t1, y_t1), xytext=(40,20), 
            textcoords='offset points', ha='center', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0.5', 
                            color='red'))
    ax1.annotate('T2: ('+str(round(x_t2, 3))+')', xy=(x_t2, y_t2), xytext=(-40,20), 
            textcoords='offset points', ha='center', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0.5', 
                            color='red'))
    ax1.annotate('Base: ('+str(round(mid, 3))+')', xy=(mid, y_mid), xytext=(-100,-20), 
            textcoords='offset points', ha='center', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0.5', 
                            color='red'))
    ##get the control function
    p = map_to_freq(t2, mid, t1, freq_lim[0], freq_lim[1])
    ##run a simulation
    if sim_bmi_fcn=='rat':
        num_t1, num_t2, num_miss = sim_bmi(data, t1, t2, mid, timeout, timeout_pause, p)
    elif sim_bmi_fcn == 'ismore':
        num_t1, num_t2, num_miss = ismore_sim_bmi(data, t1, t2, mid, timeout, timeout_pause, p)
    
    print("Simulation results:\nNumber of T1: " + str(num_t1) + "\nNumber of T2: " + str(num_t2) + "\nNumber of Misses: " + str(num_miss))
    print("Calculated T2 value is " + str(round(t2, 5)))
    print("Calculated mid value is " + str(round(mid, 5)))
    print("Calculated T1 value is " + str(round(t1, 5)))
    ##plot the control functio
    plot_cursor_func(t2, mid, t1, freq_lim[0], freq_lim[1])
    #plt.show()
    return t2, mid, t1, num_t1, num_t2, num_miss, p

def sim_bmi(baseline_data, t1, t2, midpoint, timeout, timeout_pause, p):
    data = baseline_data 
    samp_int = 100. #ms

    ##get the timeout duration in samples
    timeout_samps = int((timeout*1000.0)/samp_int)
    timeout_pause_samps = int((timeout_pause*1000.0)/samp_int)
    ##"global" variables
    num_t1 = 0
    num_t2 = 0
    num_miss = 0
    back_to_baseline = 1
    ##run through the data and simulate BMI
    i = 0
    clock = 0
    while i < (data.shape[0]-1):
        cursor = data[i]
        ##check for a target hit
        if cursor >= t1:
            num_t1+=1
            i += int(4000/samp_int)
            back_to_baseline = 0
            ##wait for a return to baseline
            while cursor >= midpoint and i < (data.shape[0]-1):
                #advance the sample
                i+=1
                ##get cursor value
                cursor = data[i]
            ##reset the clock
            clock = 0
        elif cursor <= t2:
            num_t2+=1
            i += int(4000/samp_int)
            back_to_baseline = 0
            ##wait for a return to baseline
            while cursor >= midpoint and i < (data.shape[0]-1):
                #advance the sample
                i+=1
                ##get cursor value
                cursor = data[i]
            ##reset the clock
            clock = 0
        elif clock >= timeout_samps:
            ##advance the samples for the timeout duration
            i+= timeout_pause_samps
            num_miss += 1
            ##reset the clock
            clock = 0
        else:
            ##if nothing else, advance the clock and the sample
            i+= 1
            clock+=1
    return num_t1, num_t2, num_miss

def ismore_sim_bmi(baseline_data, decoder, targets_matrix=None, session_length=0.):
    import ismore.invasive.bmi_ismoretasks as bmi_ismoretasks
    from riglib import experiment
    from features.hdf_features import SaveHDF
    from ismore.brainamp_features import SimBrainAmpData
    import datetime
    import numpy as np
    import matplotlib.pyplot as plt
    import multiprocessing as mp
    from features.blackrock_features import BlackrockBMI
    from ismore.exo_3D_visualization import Exo3DVisualizationInvasive

    targets = bmi_ismoretasks.SimBMIControlReplayFile.sleep_gen(length=100)
    plant_type = 'IsMore'
    kwargs=dict(session_length=session_length, replay_neural_features=baseline_data, decoder=decoder)
    
    if targets_matrix is not None:
        kwargs['targets_matrix']=targets_matrix
    
    Task = experiment.make(bmi_ismoretasks.SimBMIControlReplayFile, [SaveHDF])#, Exo3DVisualizationInvasive])
    task = Task(targets, plant_type=plant_type, **kwargs)
    task.run_sync()
    pnm = save_dec_enc(task)
    return pnm

def plot_cursor_func(t2, mid, t1, min_freq, max_freq):
    f, ax2 = plt.subplots()
    x = np.linspace(t2-1, t1+1, 1000)
    func = map_to_freq(t2, mid, t1, min_freq, max_freq)
    #fig, ax = plt.subplots()
    ax2.plot(t2, min_freq, 'o', color = 'r')
    ax2.plot(mid, np.floor((max_freq-min_freq)/2), 'o', color = 'r')
    ax2.plot(t1, max_freq, 'o', color = 'r')
    ax2.plot(x, func(x), '-', color = 'g')
    ax2.annotate('T1: ('+str(round(t1, 3))+')', xy=(t1, max_freq), xytext=(-20, 20), 
            textcoords='offset points', ha='center', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0.5', 
                            color='red'))
    ax2.annotate('T2: ('+str(round(t2, 3))+')', xy=(t2, min_freq), xytext=(-20,20), 
            textcoords='offset points', ha='center', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0.5', 
                            color='red'))
    ax2.annotate('Base: ('+str(round(mid, 3))+')', xy=(mid, np.floor((max_freq-min_freq)/2)), xytext=(-20,20), 
            textcoords='offset points', ha='center', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0.5', 
                            color='red'))
    ax2.set_ylabel("Feedback frequency")
    ax2.set_xlabel("Cursor value (E1-E2)")
    ax2.set_title("Cursor-frequency map", fontsize = 18)

def save_dec_enc(task, pref='sleep_sim_'):
    '''
    Summary: method to save encoder / decoder and hdf file information from task in sim_data folder
    Input param: task: task, output from arm_assist_main, or generally task object
    Input param: pref: prefix to saved file names (defaults to 'enc' for encoder)
    Output param: pkl file name used to save encoder/decoder
    '''
    #enc = task.encoder
    task.decoder.save()
    #enc.corresp_dec = task.decoder

    #Save task info
    import pickle
    ct = datetime.datetime.now()
    #pnm = '/Users/preeyakhanna/ismore/ismore_tests/sim_data/'+pref + ct.strftime("%Y%m%d_%H_%M_%S") + '.pkl'
    pnm = '/home/tecnalia/code/ismore/ismore_tests/sim_data/'+pref + ct.strftime("%m%d%y_%H%M") + '.pkl'
    pnm2 = '/Users/preeyakhanna/code/ismore/ismore_tests/sim_data/'+pref + ct.strftime("%m%d%y_%H%M") + '.pkl'
    
    try:
        pickle.dump(dict(), open(pnm,'wb'))
    except:
        pickle.dump(dict(), open(pnm2, 'wb'))
        pnm = pnm2

    #Save HDF file
    new_hdf = pnm[:-4]+'.hdf'
    import shutil
    f = open(task.h5file.name)
    f.close()

    #Wait 
    import time
    time.sleep(1.)

    #Wait after HDF cleaned up
    task.cleanup_hdf()
    import time
    time.sleep(1.)

    #Copy temp file to actual desired location
    shutil.copy(task.h5file.name, new_hdf)
    f = open(new_hdf)
    f.close()

    #Return filename
    return pnm


# Generated by Django 2.2.1 on 2019-11-13 21:58

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0002_auto_20191112_2306'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datafile',
            name='system',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.System'),
        ),
        migrations.AlterField(
            model_name='task',
            name='visible',
            field=models.BooleanField(blank=True, default=True),
        ),
    ]



from .pyeCubeStream import eCubeStream
import numpy as np
import time
from .file import *
import aopy
from riglib.source import DataSourceSystem

'''
eCube streaming sources
'''

analog_voltsperbit = 3.0517578125e-4
headstage_voltsperbit = 1.907348633e-7
TRIG_THRESH = 2.5

def multi_chan_generator(data_block, channels, downsample=1):
    for idx in range(len(channels)):
        yield (channels[idx], data_block[::downsample,idx]) # yield one channel at a time

class Broadband(DataSourceSystem):
    '''
    Wrapper class for pyecubestream compatible with using in DataSource for
    buffering neural data. Compatible with riglib.source.MultiChanDataSource
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 25000.
    dtype = np.dtype('float')

    def __init__(self, headstage=7, channels=[1]):
        '''
        Constructor for ecube.Broadband

        Inputs:
            headstages int: headstage number (1-indexed)
            channels [int array]: channel list (1-indexed)
        '''
        # Initialize the servernode-control connection
        self.conn = eCubeStream(debug=False)
        self.headstage = headstage
        self.channels = channels

    def start(self):
        print("Starting ecube streaming datasource...")

        # Remove all existing sources
        subscribed = self.conn.listadded()
        if len(subscribed[0]) > 0:
            self.conn.remove(('Headstages', self.headstage))
        if len(subscribed[1]) > 0:
            self.conn.remove(('AnalogPanel',))
        if len(subscribed[2]) > 0:
            self.conn.remove(('DigitalPanel',))

        # Add the requested headstage channels if they are available
        available = self.conn.listavailable()[0][self.headstage-1] # (headstages, analog, digital); hs are 1-indexed
        for ch in self.channels:
            if ch > available:
                raise RuntimeError('requested channel {} is not available ({} connected)'.format(
                    ch, available))
            self.conn.add(('Headstages', self.headstage, (ch, ch))) # add channels one at a time
        subscribed = self.conn.listadded() # in debug mode this prints out the added channels

        # Start streaming
        self.conn.start()

        # Start with an empty generator
        self.gen = iter(())

        # Call get once to make sure we have the right channels
        data_block = self.conn.get()
        assert data_block[1] == "Headstages"
        assert data_block[2].shape[1] == len(self.channels)
        print(f"Started streaming from ecube, packet size {data_block[2].shape[0]}")
    
    def stop(self):

        # Stop streaming
        if not self.conn.stop():
            del self.conn # try to force the streaming to end by deleting the ecube connection object
            self.conn = eCubeStream(debug=True)

        # Remove the added sources
        self.conn.remove(('Headstages', self.headstage))
        
    def get(self):
        '''data
        Retrieve a packet from the server
        '''
        try:
            return next(self.gen)
        except StopIteration:
            data_block = self.conn.get() # in the form of (time_stamp, data_source, data_content)
            # while data_block[1] != "Headstages":
            #     data_block = self.conn.get() # ideally this shouldn't happen.
            self.gen = multi_chan_generator(data_block[2], self.channels)
            return next(self.gen)

class LFP(Broadband):
    '''
    Downsample the incoming data to 1000Hz. Compatible with riglib.source.MultiChanDataSource
    '''

    update_freq = 1000.

    def get(self):
        '''data
        Retrieve a packet from the server
        '''
        try:
            return next(self.gen)
        except StopIteration:
            data_block = self.conn.get() # in the form of (time_stamp, data_source, data_content)
            self.gen = multi_chan_generator(data_block[2], self.channels, downsample=25)
            return next(self.gen)

class Digital(Broadband):
    '''
    Wrapper class for pyecubestream compatible with using in DataSource for
    buffering digital data. Compatible with riglib.source.MultiChanDataSource
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 25000.
    dtype = np.dtype('bool')

    def __init__(self, channels=[1]):
        '''
        Inputs:
            channels [int array]: channel list (1-indexed)
        '''
        # Initialize the servernode-control connection
        self.conn = eCubeStream(debug=False)
        self.channels = channels

    def start(self):
        # Remove all existing sources
        subscribed = self.conn.listadded()
        if len(subscribed[0]) > 0:
            self.conn.remove(('Headstages', self.headstage))
        if len(subscribed[1]) > 0:
            self.conn.remove(('AnalogPanel',))
        if len(subscribed[2]) > 0:
            self.conn.remove(('DigitalPanel',))

        # Add the requested headstage channels if they are available
        available = self.conn.listavailable()[2] # (headstages, analog, digital)
        for ch in self.channels:
            if ch > available:
                raise RuntimeError('requested channel {} is not available ({} connected)'.format(
                    ch, available))
            self.conn.add(('DigitalPanelAsChans', (ch, ch))) # add channels one at a time
        subscribed = self.conn.listadded() # in debug mode this prints out the added channels

        # Start streaming
        self.conn.start()

        # Start with an empty generator
        self.gen = iter(())

        # Call get once to make sure we have the right channels
        data_block = self.conn.get()
        assert data_block[1] == "DigitalPanelAsChans"
        assert data_block[2].shape[1] == len(self.channels)
        print(f"Started streaming from ecube, packet size {data_block[2].shape[0]}")
    
    def stop(self):

        # Stop streaming
        if not self.conn.stop():
            del self.conn # try to force the streaming to end by deleting the ecube connection object
            self.conn = eCubeStream(debug=True)

        # Remove the added sources
        self.conn.remove(('DigitalPanelAsChans',))

class Analog(Broadband):
    '''
    Wrapper class for pyecubestream compatible with using in DataSource for
    buffering analog data. Compatible with riglib.source.MultiChanDataSource
    '''

    def __init__(self, channels=[1]):
        '''
        Inputs:
            channels [int array]: channel list (1-indexed)
        '''
        # Initialize the servernode-control connection
        self.conn = eCubeStream(debug=False)
        self.channels = channels

    def start(self):
        # Remove all existing sources
        subscribed = self.conn.listadded()
        if len(subscribed[0]) > 0:
            self.conn.remove(('Headstages', self.headstage))
        if len(subscribed[1]) > 0:
            self.conn.remove(('AnalogPanel',))
        if len(subscribed[2]) > 0:
            self.conn.remove(('DigitalPanel',))

        # Add the requested headstage channels if they are available
        available = self.conn.listavailable()[1] # (headstages, analog, digital)
        for ch in self.channels:
            if ch > available:
                raise RuntimeError('requested channel {} is not available ({} connected)'.format(
                    ch, available))
            self.conn.add(('AnalogPanel', (ch, ch))) # add channels one at a time
        subscribed = self.conn.listadded() # in debug mode this prints out the added channels

        # Start streaming
        self.conn.start()

        # Start with an empty generator
        self.gen = iter(())

        # Call get once to make sure we have the right channels
        data_block = self.conn.get()
        assert data_block[1] == "AnalogPanel"
        assert data_block[2].shape[1] == len(self.channels)
        print(f"Started streaming from ecube, packet size {data_block[2].shape[0]}")
    
    def stop(self):

        # Stop streaming
        if not self.conn.stop():
            del self.conn # try to force the streaming to end by deleting the ecube connection object
            self.conn = eCubeStream(debug=True)

        # Remove the added sources
        self.conn.remove(('AnalogPanel',))

class File(DataSourceSystem):
    '''
    Wrapper class for pyecubestream compatible with using in DataSource for
    buffering neural data. Compatible with riglib.source.MultiChanDataSource
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 1000.
    chunksize = 728
    dtype = np.dtype('float')

    def __init__(self, ecube_bmi_filename, channels):

        # Open the file
        self.channels = channels
        zero_idx_channels = [ch-1 for ch in channels]
        self.file = aopy.data.load_ecube_data_chunked(ecube_bmi_filename, "Headstages", channels=zero_idx_channels, chunksize=self.chunksize)

    def get(self):
        '''data
        Read a "packet" worth of data from the file
        '''
        try:
            return next(self.gen)
        except (StopIteration, AttributeError):
            time.sleep(1./(25000/self.chunksize))
            try:
                data_block = next(self.file)
            except StopIteration:
                data_block = np.zeros((int(728/25),1))
            self.gen = multi_chan_generator(data_block, self.channels, downsample=25)
            return next(self.gen)

class LFP_Plus_Trigger(DataSourceSystem):
    '''
    Adds a single analog trigger as channel 0. Compatible with riglib.source.MultiChanDataSource
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 1000.
    dtype = np.dtype('float')

    def __init__(self, headstage=7, trigger_ach=0, channels=[1], trig_included_in_channels=False):
        '''
        Constructor for ecube.Broadband
        Inputs:
            headstages int: headstage number (1-indexed)
            channels [int array]: channel list (1-indexed)
        '''
        # Initialize the servernode-control connection
        self.conn = eCubeStream(debug=False)
        self.headstage = headstage
        self.trig_channel = trigger_ach
        if trig_included_in_channels:
            self.channels = channels[1:]
        else:
            self.channels = channels

    def start(self):
        print("Starting ecube streaming datasource...")

        # Remove all existing sources
        subscribed = self.conn.listadded()
        if len(subscribed[0]) > 0:
            self.conn.remove(('Headstages', self.headstage))
        if len(subscribed[1]) > 0:
            self.conn.remove(('AnalogPanel',))
        if len(subscribed[2]) > 0:
            self.conn.remove(('DigitalPanel',))

        # Add the requested headstage channels if they are available
        available = self.conn.listavailable()[0][self.headstage-1] # (headstages, analog, digital); hs are 1-indexed
        for ch in self.channels:
            if ch > available:
                raise RuntimeError('requested channel {} is not available ({} connected)'.format(
                    ch, available))
            self.conn.add(('Headstages', self.headstage, (ch, ch))) # add channels one at a time

        # Add the digital panel for triggering
        self.conn.add(('AnalogPanel', (self.trig_channel, self.trig_channel)))
        subscribed = self.conn.listadded() # in debug mode this prints out the added channels

        # Start streaming
        self.conn.start()

        # Start with an empty generator for the headstage channels
        self.gen = iter(())

    
    def stop(self):

        # Stop streaming
        if not self.conn.stop():
            del self.conn # try to force the streaming to end by deleting the ecube connection object
            self.conn = eCubeStream(debug=True)

        # Remove the added sources
        self.conn.remove(('Headstages', self.headstage))
        self.conn.remove(('AnalogPanel',))
        
    def get(self):
        '''data
        Retrieve a packet from the server
        '''
        try:
            return next(self.gen)
        except StopIteration:
            data_block = self.conn.get() # in the form of (time_stamp, data_source, data_content)
            if data_block[1] == "Headstages":
                self.gen = multi_chan_generator(data_block[2], self.channels, downsample=25)
                return next(self.gen)
            else:
                return 0, data_block[2][::25] # The trigger data
    
    

class LFP_Plus_Trigger_File(DataSourceSystem):
    '''
    Adds a single analog trigger as channel 0, but reading from file. Compatible with riglib.source.MultiChanDataSource
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 1000.
    chunksize = 728
    dtype = np.dtype('float')

    def __init__(self, channels, ecube_bmi_filename=None, trig_channel=0, trig_included_in_channels=False):
        
        if not ecube_bmi_filename:
            ecube_bmi_filename = "/Users/leoscholl/2023-10-10_BMI3D_te11503"

        # Open the file
        self.trig_channel = trig_channel
        if trig_included_in_channels:
            self.channels = channels[1:]
        else:
            self.channels = channels
        zero_idx_channels = [ch-1 for ch in self.channels]
        self.file = aopy.data.load_ecube_data_chunked(ecube_bmi_filename, "Headstages", channels=zero_idx_channels, chunksize=self.chunksize)
        self.trig_file = aopy.data.load_ecube_data_chunked(ecube_bmi_filename, "AnalogPanel", channels=[self.trig_channel], chunksize=self.chunksize)
        self.trig_flag = True
        
    def get(self):
        '''data
        Read a "packet" worth of data from the file
        '''
        try:
            return next(self.gen)
        except (StopIteration, AttributeError):
            time.sleep(1./(25000/self.chunksize))
            if self.trig_flag:
                try:
                    trig_chunk = next(self.trig_file)[::25]
                except StopIteration:
                    trig_chunk = np.zeros((int(728/25),1))
                self.trig_flag = False
                return 0, np.squeeze(trig_chunk)
            try:
                data_block = next(self.file)
                self.trig_flag = True
            except StopIteration:
                data_block = np.zeros((int(728/25),len(self.channels)))
            self.gen = multi_chan_generator(data_block, self.channels, downsample=25)
            return next(self.gen)
        
def map_channels_for_multisource(headstage_channels=[], analog_channels=[], digital_channels=[]):
    '''
    Utility to convert the channel numbers to the format used by the MultiSource class

    Args:
        headstage_channels (list): list of 1-indexed headstage channels
        analog_channels (list): list of 1-indexed analog channels
        digital_channels (list): list of 1-indexed digital channels

    Returns:
        np.array: channel numbers in the format used by the MultiSource class
    '''
    return np.concatenate((analog_channels, np.array(digital_channels)+32, np.array(headstage_channels)+96))

class MultiSource(DataSourceSystem):
    '''
    Adds all Analog, Digital and Broadband channels. Compatible with riglib.source.MultiChanDataSource.
    Because of the way the system is set up, Analog channels map from 1-32, Digital channels 
    from 33-96, and Broadband channels from 97- onwards. Use the mapping utility to convert channels.
    Not tested for use with BMI, as the latency is likely higher than streaming only broadband data.
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 25000.
    dtype = np.dtype('float')

    @classmethod
    def pre_init(cls, headstage=7, n_digital_channels=64, n_analog_channels=32, n_headstage_channels=256):
        '''
        Set up servernode with all the possible channels you want to use
        '''
        conn = eCubeStream(autoshutdown=False, debug=True)

        # Remove all existing sources
        subscribed = conn.listadded()
        if len(subscribed[0]) > 0:
            conn.remove(('Headstages', headstage))
        if len(subscribed[1]) > 0:
            conn.remove(('AnalogPanel',))
        if len(subscribed[2]) > 0:
            conn.remove(('DigitalPanel',))

        # Add all available channels
        conn.add(('Headstages', headstage, (1, n_headstage_channels)))
        conn.add(('AnalogPanel', (1, n_analog_channels)))
        conn.add(('DigitalPanelAsChans', (1, n_digital_channels)))

        subscribed = conn.listadded() # in debug mode this prints out the added channels

    def __init__(self, headstage=7, channels=[]):
        '''
        Inputs:
            headstages int: headstage number (1-indexed)
            channels [int array]: channel list (1-indexed) where channels 1-32 are analog, 
                33-96 are digital, and 97- are broadband channels
        '''
        self.conn = eCubeStream(autoshutdown=False, debug=True)
        channels = np.array(channels)
        self.analog_channels = (channels[channels <= 32]).astype(int).tolist()
        self.digital_channels = (channels[(channels > 32) & (channels <= 96)] - 32).astype(int).tolist()
        self.headstage = headstage
        self.headstage_channels = (channels[channels > 96] - 96).astype(int).tolist()
        headstage_sub, analog_sub, digital_sub = self.conn.listadded()
        assert np.all(headstage_sub[0] == self.headstage) # can't handle multiple headstages yet
        self.headstage_idx = np.isin(headstage_sub[1], self.headstage_channels)
        self.analog_idx = np.isin(analog_sub, self.analog_channels)
        self.digital_idx = np.isin(digital_sub, self.digital_channels)
        print('digital channels matched', np.sum(self.digital_idx))

    def start(self):
        print("Starting ecube streaming datasource...")
        self.conn.start()

        # Start with an empty generator
        self.gen = iter(())
        print('...started!')
    
    def stop(self):

        # Stop streaming
        if not self.conn.stop():
            del self.conn # try to force the streaming to end by deleting the ecube connection object
        
    def get(self):
        '''
        Retrieve a packet from the server
        '''
        try:
            return next(self.gen)
        except StopIteration:
            data_block = self.conn.get() # in the form of (time_stamp, data_source, data_content)
            if data_block[1] == "AnalogPanel":
                if np.sum(self.analog_idx):
                    self.gen = multi_chan_generator(data_block[2][:,self.analog_idx], self.analog_channels)
                else:
                    return self.get()
            elif data_block[1] == "DigitalPanel": # If you subscribe to all digital channels this happens
                if np.sum(self.digital_idx):
                    digital_data = aopy.utils.convert_digital_to_channels(data_block[2])[:,self.digital_channels]
                    self.gen = multi_chan_generator(digital_data, [ch+32 for ch in self.digital_channels])
                else:
                    return self.get()
            elif data_block[1] == "DigitalPanelAsChans":
                if np.sum(self.digital_idx):
                    self.gen = multi_chan_generator(data_block[2][:,self.digital_idx], [ch+32 for ch in self.digital_channels])
                else:
                    return self.get()
            elif data_block[1] == "Headstages":
                if np.sum(self.headstage_idx):
                    self.gen = multi_chan_generator(data_block[2][:,self.headstage_idx], [ch+96 for ch in self.headstage_channels])
                else:
                    return self.get()
            else:
                return self.get()
            return next(self.gen)
    
class MultiSource_File(DataSourceSystem):
    '''
    Adds digital, analog, and headstage data, but reading from file. Compatible with riglib.source.MultiChanDataSource
    '''
    # Required by DataSourceSystem: update_freq and dtype (see make() below)
    update_freq = 25000.
    chunksize = 728
    dtype = np.dtype('float')

    def __init__(self, channels, ecube_bmi_filename=None):
        
        if not ecube_bmi_filename:
            # ecube_bmi_filename = "/media/moor-data/raw/ecube/2024-03-04_BMI3D_te14929" # laser
            ecube_bmi_filename = "/media/moor-data/raw/ecube/2024-03-06_BMI3D_te15031" # mc
        # Open the file
        channels = np.array(channels)
        self.analog_channels = (channels[channels <= 32] - 1).astype('int')
        self.digital_channels = (channels[(channels > 32) & (channels <= 96)] - 33).astype('int')
        self.headstage_channels = (channels[channels > 96] - 97).astype('int')

        self.headstage_file = aopy.data.load_ecube_data_chunked(ecube_bmi_filename, "Headstages", channels=self.headstage_channels, chunksize=self.chunksize)
        self.analog_file = aopy.data.load_ecube_data_chunked(ecube_bmi_filename, "AnalogPanel", channels=self.analog_channels, chunksize=self.chunksize)
        self.digital_file = aopy.data.load_ecube_data_chunked(ecube_bmi_filename, "DigitalPanel", chunksize=self.chunksize)
        self.flag = "Headstages"
        self.gen = iter(())
        self.t0 = time.perf_counter()
        
    def _next_gen(self, file, channels):
        try:
            data_block = next(file)
        except StopIteration:
            data_block = np.zeros((int(self.chunksize),len(channels)))
        self.gen = multi_chan_generator(data_block, channels)
        return next(self.gen)

    def get(self):
        '''
        Read a "packet" worth of data from one of the files
        '''
        try:
            return next(self.gen)
        except (StopIteration, AttributeError):
            if self.flag == "Headstages" and len(self.headstage_channels) > 0:
                self.flag = "DigitalPanelAsChans"
                return self._next_gen(self.headstage_file, self.headstage_channels+97)
            elif self.flag == "Headstages":
                self.flag = "DigitalPanelAsChans"
            
            if self.flag == "DigitalPanelAsChans" and len(self.digital_channels) > 0:
                self.flag = "AnalogPanel"
                try:
                    # Special consideration for digital panel to convert bits to channels
                    data_block = next(self.digital_file)
                    data_block = aopy.utils.convert_digital_to_channels(data_block)[:,self.digital_channels]
                except StopIteration:
                    data_block = np.zeros((int(self.chunksize),len(self.digital_channels)))
                self.gen = multi_chan_generator(data_block, self.digital_channels+33)
                return next(self.gen)
            elif self.flag == "DigitalPanelAsChans":
                self.flag = "AnalogPanel"
            
            if self.flag == "AnalogPanel" and len(self.analog_channels) > 0:
                self.flag = "Headstages"
                while time.perf_counter() - self.t0 < 1./(25000/self.chunksize):
                    pass
                self.t0 = time.perf_counter()
                return self._next_gen(self.analog_file, self.analog_channels+1)
            elif self.flag == "AnalogPanel":
                self.flag = "Headstages"
                return self.get() # Try again
            

class LFP_Blanking(LFP_Plus_Trigger):
    '''
    Blanks LFP data in the interval when the trigger is on. Compatible with riglib.source.MultiChanDataSource.

    Limitation: only blanks one chunk at a time, so smaller chunks are better! But -- the analog panel stream
    arrives in more frequent 22-sample packets; once the headstage packets are that size then there is no 
    benefit to making the hs packets any smaller! 
    
    Warning: untested! Use at your own risk.
    '''

    blanking = 0
    headstage_buffer = []
    trig_buffer = []
    
    def get(self):
        '''data
        Retrieve a packet from the server
        '''
        try:
            return next(self.gen)        
        except StopIteration:
            data_block = self.conn.get() # in the form of (time_stamp, data_source, data_content)
            while data_block[1] != "Headstages": # new packet of trigger data
                self.trig_buffer = np.concatenate((self.trig_buffer, np.squeeze(data_block[2])), axis=0)
                data_block = self.conn.get() # in the form of (time_stamp, data_source, data_content)

            if len(self.trig_buffer) == 0:
                self.blanking = 1
            else:
                trig_chunk_volts = analog_voltsperbit*np.array(self.trig_buffer)
            if self.blanking:
                self.blanking -= 1
            elif (trig_chunk_volts[-1] > TRIG_THRESH and trig_chunk_volts[0] < TRIG_THRESH):
                self.blanking = 1
            elif (trig_chunk_volts[-1] < TRIG_THRESH and trig_chunk_volts[0] > TRIG_THRESH):
                self.blanking = 1
            self.trig_buffer = [] # reset the trigger accumulation

            if self.blanking:
                self.gen = multi_chan_generator(self.headstage_buffer, self.channels, downsample=25)
            else:
                self.headstage_buffer = data_block[2].copy() # save this good hs data for later if blanking is needed
                self.gen = multi_chan_generator(data_block[2], self.channels, downsample=25)
            return next(self.gen)


class LFP_Blanking_File(LFP_Plus_Trigger_File):
    '''
    Blanks LFP data in the interval when the trigger is on. Compatible with riglib.source.MultiChanDataSource
    '''

    blanking = 0
    headstage_buffer = []
    
    def get(self):
        '''data
        Read a "packet" worth of data from the file
        '''
        try:
            return next(self.gen)        
        except (StopIteration, AttributeError):
            time.sleep(1./(25000/self.chunksize))
            
            # Get a chunk of trigger data
            try:
                trig_chunk = next(self.trig_file)
            except StopIteration:
                trig_chunk = np.zeros((self.chunksize,))
            trig_chunk_volts = analog_voltsperbit*trig_chunk
            if self.blanking:
                self.blanking -= 1
            elif (trig_chunk_volts[-1] > TRIG_THRESH and trig_chunk_volts[0] < TRIG_THRESH):
                self.blanking = 1
            elif (trig_chunk_volts[-1] < TRIG_THRESH and trig_chunk_volts[0] > TRIG_THRESH):
                self.blanking = 1
                  
            # Get a chunk of hs data
            try:
                data_block = next(self.file)
            except StopIteration:
                data_block = np.zeros((self.chunksize,len(self.channels)))
                
            # Make the new generator and return the first channel
            if self.blanking:
                data_block = self.headstage_buffer.copy() # copy the last good headstage data for this channel
            else:
                self.headstage_buffer = data_block.copy() # save this good hs data for later if blanking is needed
            self.gen = multi_chan_generator(data_block, self.channels, downsample=25)
            return next(self.gen)
    
def make_source_class(cls, trigger_ach):
    
    def init(self, **kwargs):
        super(self.__class__, self).__init__(trigger_ach=trigger_ach, **kwargs)
        
    return type(cls.__name__, (cls,), dict(__init__=init))

# e.g.
# cls = make_source_class(LFP_Plus_Trigger, 17) # laser ch 2



"""
Trajectory primitives.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from . import transform as T

def jtraj(q0, q1, tv, qd0=None, qd1=None):
    """
    Compute a joint space trajectory between points C{q0} and C{q1}.
    The number of points is the length of the given time vector C{tv}.  If
    {tv} is a scalar it is taken as the number of points.
    
    A 7th order polynomial is used with default zero boundary conditions for
    velocity and acceleration.  Non-zero boundary velocities can be
    optionally specified as C{qd0} and C{qd1}.
    
    As well as the trajectory, M{q{t}}, its first and second derivatives
    M{qd(t)} and M{qdd(t)} are also computed.  All three are returned as a tuple.
    Each of these is an M{m x n} matrix, with one row per time step, and
    one column per joint parameter.

    @type q0: m-vector
    @param q0: initial state
    @type q1: m-vector
    @param q1: final state
    @type tv: n-vector or scalar
    @param tv: time step vector or number of steps
    @type qd0: m-vector
    @param qd0: initial velocity (default 0)
    @type qd1: m-vector
    @param qd1: final velocity (default 0)
    @rtype: tuple
    @return: (q, qd, qdd), a tuple of M{m x n} matrices
    @see: L{ctraj}

    """

    if isinstance(tv,(int,int32,float,float64)):
        tscal = float(1)
        t = mat(range(0,tv)).T/(tv-1.) # Normalized time from 0 -> 1
    else:
        tv = arg2array(tv);
        tscal = float(max(tv))
        t = mat(tv).T / tscal
    
    q0 = arg2array(q0)
    q1 = arg2array(q1)
    
    if qd0 == None:
        qd0 = zeros((shape(q0)))
    else:
        qd0 = arg2array(qd0);
    if qd1 == None:
        qd1 = zeros((shape(q1)))
    else:
        qd1 = arg2array(qd1)
    
    # compute the polynomial coefficients
    A = 6*(q1 - q0) - 3*(qd1 + qd0)*tscal
    B = -15*(q1 - q0) + (8*qd0 + 7*qd1)*tscal
    C = 10*(q1 - q0) - (6*qd0 + 4*qd1)*tscal
    E = qd0*tscal # as the t vector has been normalized
    F = q0

    tt = concatenate((power(t,5),power(t,4),power(t,3),power(t,2),t,ones(shape(t))),1)
    c = vstack((A, B, C, zeros(shape(A)), E, F))
    qt = tt * c

    # compute velocity
    c = vstack((zeros(shape(A)),5*A,4*B,3*C,zeros(shape(A)),E))
    qdt = tt * c / tscal


    # compute acceleration
    c = vstack((zeros(shape(A)),zeros(shape(A)),20*A,12*B,6*C,zeros(shape(A))))
    qddt = tt * c / (tscal**2)

    return qt,qdt,qddt


def ctraj(t0, t1, r):
    """
    Compute a Cartesian trajectory between poses C{t0} and C{t1}.
    The number of points is the length of the path distance vector C{r}.
    Each element of C{r} gives the distance along the path, and the 
    must be in the range [0 1].

    If {r} is a scalar it is taken as the number of points, and the points 
    are equally spaced between C{t0} and C{t1}.

    The trajectory is a list of transform matrices.
    
    @type t0: homogeneous transform
    @param t0: initial pose
    @rtype: list of M{4x4} matrices
    @return: Cartesian trajectory
    @see: L{trinterp}, L{jtraj}
    """

    if isinstance(r,(int,int32,float,float64)):
        i = mat(range(1,r+1))
        r = (i-1.)/(r-1)
    else:
        r = arg2array(r);

    if any(r>1) or any(r<0):
        raise 'path position values (R) must 0<=R<=1'
    traj = []
    for s in r.T:
        traj.append( T.trinterp(t0, t1, float(s)) )
    return traj
    




'''
Classes for extracting "decodable features" from various types of neural signal sources. 
Examples include spike rate estimation, LFP power, and EMG amplitude.
'''
import numpy as np
from scipy.signal import butter, lfilter
import os
import nitime.algorithms as tsa
import aopy

class FeatureExtractor(object):
    '''
    Parent of all feature extractors, used only for interfacing/type-checking.
    Feature extractors are objects tha gets the data that it needs (e.g., spike timestamps, LFP voltages, etc.) 
    from the neural data source object and extracts features from it
    '''
    @classmethod
    def extract_from_file(cls, *args, **kwargs):
        raise NotImplementedError


class DummyExtractor(FeatureExtractor):
    '''
    An extractor which does nothing. Used for tasks which are only pretending to be BMI tasks, e.g., visual feedback tasks
    '''
    feature_type = 'obs'
    feature_dtype = [('obs', 'f8', (1,))]

    def __call__(self, *args, **kwargs):
        return dict(obs=np.array([[np.nan]]))

class DirectObsExtractor(FeatureExtractor):
    '''
    Doesn't 'extract' anything, just passes on the input
    '''
    feature_type = 'obs'

    def __init__(self, source):
        test = source.get()
        self.feature_dtype = [('obs', source.source.dtype, np.shape(test))]
        self.source = source

    def __call__(self, *args, **kwargs):
        return dict(obs=self.source.get())

class BinnedSpikeCountsExtractor(FeatureExtractor):
    '''
    Extracts spike counts from spike timestamps separated into rectangular window. 
    This extractor is (currently) the main type of feature extractor in intracortical BMIs
    '''
    feature_type = 'spike_counts'

    def __init__(self, source, n_subbins=1, units=[]):
        '''
        Constructor for BinnedSpikeCountsExtractor

        Parameters
        ----------
        source: DataSource instance
            Source must implement a '.get()' function which returns the appropriate data 
            (appropriateness will change depending on the source)
        n_subbins: int, optional, default=1
            Number of bins into which to divide the observed spike counts 
        units: np.ndarray of shape (N, 2), optional, default=[]
            Units which need spike binning. Each row of the array corresponds to (channel, unit). By default no units will be binned.

        Returns
        -------
        BinnedSpikeCountsExtractor instance
        '''
        self.feature_dtype = [('spike_counts', 'u4', (len(units), n_subbins)), ('bin_edges', 'f8', 2)]

        self.source = source
        self.n_subbins = n_subbins
        self.units = units

        extractor_kwargs = dict()
        extractor_kwargs['n_subbins'] = self.n_subbins
        extractor_kwargs['units']     = self.units
        self.extractor_kwargs = extractor_kwargs
        self.last_get_spike_counts_time = 0

    def set_n_subbins(self, n_subbins):
        '''
        Alter the # of subbins without changing the extractor kwargs of a decoder

        Parameters
        ----------
        n_subbins : int 
            Number of bins into which to divide the observed spike counts 

        Returns
        -------
        None
        '''
        self.n_subbins = n_subbins
        self.extractor_kwargs['n_subbins'] = n_subbins
        self.feature_dtype = [('spike_counts', 'u4', (len(self.units), n_subbins)), ('bin_edges', 'f8', 2)]

    def get_spike_ts(self, *args, **kwargs):
        '''
        Get the spike timestamps from the neural data source. This function has no type checking, 
        i.e., it is assumed that the Extractor object was created with the proper source

        Parameters
        ----------
        None are needed (args and kwargs are ignored)

        Returns
        -------
        numpy record array
            Spike timestamps in the format of riglib.plexon.Spikes.dtype
        '''
        return self.source.get()

    def get_bin_edges(self, ts):
        '''
        Determine the first and last spike timestamps to allow HDF files 
        created by the BMI to be semi-synchronized with the neural data file

        Parameters
        ----------
        ts : numpy record array
            Must have field 'ts' of spike timestamps in seconds

        Returns
        -------
        np.ndarray of shape (2,)
            The smallest and largest timestamps corresponding to the current feature; 
            useful for rough synchronization of the BMI event loop with the neural recording system.
        '''
        if len(ts) == 0:
            bin_edges = np.array([np.nan, np.nan])
        else:
            min_ind = np.argmin(ts['ts'])
            max_ind = np.argmax(ts['ts'])
            bin_edges = np.array([ts[min_ind]['ts'], ts[max_ind]['ts']])

    @classmethod
    def bin_spikes(cls, ts, units, max_units_per_channel=13):
        '''
        Count up the number of BMI spikes in a list of spike timestamps.

        Parameters
        ----------
        ts : numpy record array
            Must have field 'ts' of spike timestamps in seconds
        units : np.ndarray of shape (N, 2)
            Each row corresponds to the channel index (typically the electrode number) and 
            the unit index (an index to differentiate the possibly many units on the same electrode). These are 
            the units used in the BMI.
        max_units_per_channel : int, optional, default=13
            This int is used to map from a (channel, unit) index to a single 'unit_ind' 
            for faster binning of spike timestamps. Just set to a large number.

        Returns
        -------
        np.ndarray of shape (N, 1)
            Column vector of counts of spike events for each of the N units.
        '''
        unit_inds = units[:,0]*max_units_per_channel + units[:,1]
        edges = np.sort(np.hstack([unit_inds - 0.5, unit_inds + 0.5]))

        spiking_unit_inds = ts['chan']*max_units_per_channel + ts['unit']
        counts, _ = np.histogram(spiking_unit_inds, edges)
        return counts[::2]

    def __call__(self, start_time, *args, **kwargs):
        '''
        Main function to retreive new spike data and bin the counts

        Parameters
        ----------
        start_time : float 
            Absolute time from the task event loop. This is used only to subdivide 
            the spike timestamps into multiple bins, if desired (if the 'n_subbins' attribute is > 1)
        *args, **kwargs : optional positional/keyword arguments
            These are passed to the source, or ignored (not needed for this extractor).

        Returns
        -------
        dict
            Extracted features to be saved in the task. 
        '''
        ts = self.get_spike_ts(*args, **kwargs)
        if len(ts) == 0:
            counts = np.zeros([len(self.units), self.n_subbins])
        elif self.n_subbins > 1:
            subbin_edges = np.linspace(self.last_get_spike_counts_time, start_time, self.n_subbins+1)

            # Decrease the first subbin index to include any spikes that were
            # delayed in getting to the task layer due to threading issues
            # An acceptable delay is 1 sec or less. Realistically, most delays should be
            # on the millisecond order
            subbin_edges[0] -= 1
            subbin_inds = np.digitize(ts['arrival_ts'], subbin_edges)
            counts = np.vstack([self.bin_spikes(ts[subbin_inds == k], self.units) for k in range(1, self.n_subbins+1)]).T
        else:
            counts = self.bin_spikes(ts, self.units).reshape(-1, 1)

        counts = np.array(counts, dtype=np.uint32)
        bin_edges = self.get_bin_edges(ts)
        self.last_get_spike_counts_time = start_time

        return dict(spike_counts=counts, bin_edges=bin_edges)

    @classmethod
    def extract_from_file(cls, files, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
        '''
        Compute binned spike count features

        Parameters
        ----------
        files : dict
            Data files used to train the decoder. Should contain exactly one type of neural data file (e.g., Plexon, Blackrock, TDT)
        neurows: np.ndarray of shape (T,)
            Timestamps in the plexon time reference corresponding to bin boundaries
        binlen: float
            Length of time over which to sum spikes from the specified cells
        units: np.ndarray of shape (N, 2)
            List of units that the decoder will be trained on. The first column specifies the electrode number and the second specifies the unit on the electrode
        extractor_kwargs: dict 
            Any additional parameters to be passed to the feature extractor. This function is agnostic to the actual extractor utilized
        strobe_rate: 60.0
            The rate at which the task sends the sync pulse to the plx file

        Returns
        -------
        spike_counts : np.ndarray of shape (N, T)
            Spike counts binned over the length of the datafile.
        units : np.ndarray of shape (N, 2)
            Each row corresponds to the channel index (typically the electrode number) and 
            the unit index (an index to differentiate the possibly many units on the same electrode). These are 
            the units used in the BMI.
        extractor_kwargs : dict
            Parameters used to instantiate the feature extractor, to be stored 
            along with the trained decoder so that the exact same feature extractor can be re-created at runtime.
        '''
        if 'plexon' in files:
            from plexon import plexfile
            plx = plexfile.openFile(files['plexon'].encode('utf-8'))
            # interpolate between the rows to 180 Hz
            if binlen < 1./strobe_rate:
                interp_rows = []
                neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
                for r1, r2 in zip(neurows[:-1], neurows[1:]):
                    interp_rows += list(np.linspace(r1, r2, 4)[1:])
                interp_rows = np.array(interp_rows)
            else:
                step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
                interp_rows = neurows[::step]
            print(('step: ', step))
            from plexon import psth
            spike_bin_fn = psth.SpikeBin(units, binlen)
            spike_counts = np.array(list(plx.spikes.bin(interp_rows, spike_bin_fn)))


            # discard units that never fired at all
            discard_zero_units = extractor_kwargs.pop('discard_zero_units', True)
            if discard_zero_units:
                unit_inds, = np.nonzero(np.sum(spike_counts, axis=0))
                units = units[unit_inds,:]
                spike_counts = spike_counts[:, unit_inds]
                extractor_kwargs['units'] = units

            return spike_counts, units, extractor_kwargs

        elif 'blackrock' in files:
            nev_fname = [name for name in files['blackrock'] if '.nev' in name][0]  # only one of them
            nev_hdf_fname = [name for name in files['blackrock'] if '.nev' in name and name[-4:]=='.hdf']
            nsx_fnames = [name for name in files['blackrock'] if '.ns' in name]            
            # interpolate between the rows to 180 Hz
            if binlen < 1./strobe_rate:
                interp_rows = []
                neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
                for r1, r2 in zip(neurows[:-1], neurows[1:]):
                    interp_rows += list(np.linspace(r1, r2, 4)[1:])
                interp_rows = np.array(interp_rows)
            else:
                step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
                interp_rows = neurows[::step]

            
            
            if len(nev_hdf_fname) == 0:
                nev_hdf_fname = nev_fname + '.hdf'
                
                if not os.path.isfile(nev_hdf_fname):
                    # convert .nev file to hdf file using Blackrock's n2h5 utility
                    subprocess.call(['n2h5', nev_fname, nev_hdf_fname])
            else:
                nev_hdf_fname = nev_hdf_fname[0]

            try:
                nev_hdf = h5py.File(nev_hdf_fname, 'r')
                open_method = 1
            except:
                import tables
                nev_hdf = tables.openFile(nev_hdf_fname)
                open_method = 2
                #print 'open method 2'

            n_bins = len(interp_rows)
            n_units = units.shape[0]
            spike_counts = np.zeros((n_bins, n_units))

            for i in range(n_units):
                chan = units[i, 0]

                # 1-based numbering (comes from web interface)
                unit = units[i, 1]  

                chan_str = str(chan).zfill(5)
                path = 'channel/channel%s/spike_set' % chan_str

                if open_method == 1:    
                    ts = nev_hdf.get(path).value['TimeStamp']
                    # the units corresponding to each timestamp in ts
                    # 0-based numbering (comes from .nev file), so add 1
                    units_ts = nev_hdf.get(path).value['Unit']
                
                elif open_method == 2:
                    try:
                        grp = nev_hdf.getNode('/'+path)
                        ts = grp[:]['TimeStamp']
                        units_ts = grp[:]['Unit']
                    except:
                        print(('no spikes recorded on channel: ', chan_str, ': adding zeros'))
                        ts = []
                        units_ts = []


                # get the ts for this unit, in units of secs
                fs = 30000.
                ts = [t/fs for idx, (t, u_t) in enumerate(zip(ts, units_ts)) if u_t == unit]

                # insert value interp_rows[0]-step to beginning of interp_rows array
                interp_rows_ = np.insert(interp_rows, 0, interp_rows[0]-step)

                # use ts to fill in the spike_counts that corresponds to unit i
                spike_counts[:, i] = np.histogram(ts, interp_rows_)[0]


            # discard units that never fired at all
            if 'keep_zero_units' in extractor_kwargs:
                print('keeping zero firing units')
            else:
                unit_inds, = np.nonzero(np.sum(spike_counts, axis=0))
                units = units[unit_inds,:]
                spike_counts = spike_counts[:, unit_inds]
            
            extractor_kwargs['units'] = units

            return spike_counts, units, extractor_kwargs  
                 
        elif 'tdt' in files:
            raise NotImplementedError     

        elif 'ecube' in files:
            raise NotImplementedError

# bands should be a list of tuples representing ranges
#   e.g., bands = [(0, 10), (10, 20), (130, 140)] for 0-10, 10-20, and 130-140 Hz
start = 0
end   = 150
step  = 10
default_bands = []
for freq in range(start, end, step):
    default_bands.append((freq, freq+step))

class rms_emg(object):
    '''compute RMS for emg per Si Jia code'''

    feature_type = 'emg_amplitude'

    def __init__(self, source, channels=[], win_len=0.2, fs=2048, **kwargs):
        '''
        Constructor for emg rms extractor, which extracts emg rms

        Parameters
        ----------
        source : riglib.source.Source object
            Object which yields new data when its 'get' method is called
        channels : list 
            EMG electrode indices to use for feature extraction

        Returns
        -------
        rms extractor instance
        '''
        self.source = source
        self.channels = channels
        if source is not None:
            self.fs = source.source.update_freq
        else:
            self.fs = fs

        self.win_len = win_len

        extractor_kwargs = dict()

        extractor_kwargs['win_len']  = self.win_len
        extractor_kwargs['fs']       = self.fs
        self.n_pts = int(self.win_len * self.fs)

        
        extractor_kwargs['channels'] = self.channels
        extractor_kwargs['fs']       = self.fs
        
        self.feature_dtype = ('emg_amplitude', 'f8', (len(channels),1))

    def get_cont_samples(self, *args, **kwargs):
        '''
        Retreives the last n_pts number of samples for each emg channel from the data 'source'

        Parameters
        ----------
        *args, **kwargs : optional arguments
            Ignored for this extractor (not necessary)

        Returns
        -------
        np.ndarray of shape ???
        '''
        return self.source.get(self.n_pts, self.channels)

    def extract_features(self, cont_samples):
        '''
        Extract spectral features from a block of time series samples

        Parameters
        ----------
        cont_samples : np.ndarray of shape (n_channels, n_samples)
            Raw voltage time series (one per channel) from which to extract spectral features 

        Returns
        -------
        lfp_power : np.ndarray of shape (n_channels * n_features, 1)
            Multi-band power estimates for each channel, for each band specified when the feature extractor was instantiated.
        '''
        assert int(self.win_len * self.fs) == cont_samples.shape[1]

        emg_amplitude = np.zeros([1,cont_samples.shape[0],])
        for i,row in enumerate(cont_samples):
            emg_amplitude[0,i] = np.sqrt(np.sum([i ** 2 for i in row])/cont_samples.shape[0])
        return emg_amplitude.T

    def __call__(self, start_time, *args, **kwargs):
        '''
        Parameters
        ----------
        start_time : float 
            Absolute time from the task event loop. This is unused by LFP extractors in their current implementation
            and only passed in to ensure that function signatures are the same across extractors.
        *args, **kwargs : optional positional/keyword arguments
            These are passed to the source, or ignored (not needed for this extractor).

        Returns
        -------
        dict
            Extracted features to be saved in the task.         
        '''
        cont_samples = self.get_cont_samples(*args, **kwargs)  # dims of channels x time
        emg_amplitude = self.extract_features(cont_samples)

        return dict(emg_amplitude=emg_amplitude)


class LFPMTMPowerExtractor(object):
    '''
    Computes log power of the LFP in different frequency bands (for each 
    channel) in freq-domain using the multi-taper method.
    '''

    feature_type = 'lfp_power'

    def __init__(self, source, channels=[], bands=default_bands, win_len=0.2, NW=3, fs=1000, **kwargs):
        '''
        Constructor for LFPMTMPowerExtractor, which extracts LFP power using the multi-taper method

        Parameters
        ----------
        source : riglib.source.Source object
            Object which yields new data when its 'get' method is called
        channels : list 
            LFP electrode indices to use for feature extraction
        bands : list of tuples
            Each tuple defines a frequency band of interest as (start frequency, end frequency) 

        Returns
        -------
        LFPMTMPowerExtractor instance
        '''
        self.source = source
        self.channels = channels
        self.bands = bands
        self.win_len = win_len
        self.NW = NW
        if source is not None:
            self.fs = source.source.update_freq
        else:
            self.fs = fs

        extractor_kwargs = dict()
        extractor_kwargs['channels'] = self.channels
        extractor_kwargs['bands']    = self.bands
        extractor_kwargs['win_len']  = self.win_len
        extractor_kwargs['NW']       = self.NW
        extractor_kwargs['fs']       = self.fs
        self.npk = aopy.precondition.convert_taper_parameters(win_len, NW/win_len)
   
        extractor_kwargs['no_log']  = 'no_log' in kwargs and kwargs['no_log']==True #remove log calculation
        extractor_kwargs['no_mean'] = 'no_mean' in kwargs and kwargs['no_mean']==True #r
        self.extractor_kwargs = extractor_kwargs

        self.n_pts = int(self.win_len * self.fs)
        self.nfft = 2**int(np.ceil(np.log2(self.n_pts)))  # nextpow2(self.n_pts)
        fft_freqs = np.arange(0., fs, float(fs)/self.nfft)[:int(self.nfft/2) + 1]
        self.fft_inds = dict()
        for band_idx, band in enumerate(bands):
            self.fft_inds[band_idx] = [freq_idx for freq_idx, freq in enumerate(fft_freqs) if band[0] <= freq < band[1]]

        extractor_kwargs['fft_inds']       = self.fft_inds
        extractor_kwargs['fft_freqs']      = fft_freqs
        
        self.epsilon = 1e-9
        if extractor_kwargs['no_mean']: #Used in lfp 1D control task
            self.feature_dtype = ('lfp_power', 'f8', (len(channels)*len(fft_freqs), 1))
        else:
            self.feature_dtype = ('lfp_power', 'f8', (len(channels)*len(bands), 1))

    def get_cont_samples(self, *args, **kwargs):
        '''
        Retreives the last n_pts number of samples for each LPF channel from the neural data 'source'

        Parameters
        ----------
        *args, **kwargs : optional arguments
            Ignored for this extractor (not necessary)

        Returns
        -------
        np.ndarray of shape ???
        '''
        return self.source.get(self.n_pts, self.channels)

    # def extract_features(self, cont_samples):
    #     '''
    #     Extract spectral features from a block of time series samples

    #     Parameters
    #     ----------
    #     cont_samples : np.ndarray of shape (n_channels, n_samples)
    #         Raw voltage time series (one per channel) from which to extract spectral features 

    #     Returns
    #     -------
    #     lfp_power : np.ndarray of shape (n_channels * n_features, 1)
    #         Multi-band power estimates for each channel, for each band specified when the feature extractor was instantiated.
    #     '''
    #     psd_est = tsa.multi_taper_psd(cont_samples, Fs=self.fs, NW=self.NW, jackknife=False, low_bias=True, NFFT=self.nfft)[1]
        
    #     if ('no_mean' in self.extractor_kwargs) and (self.extractor_kwargs['no_mean'] is True):
    #         return psd_est.reshape(psd_est.shape[0]*psd_est.shape[1], 1)

    #     else:
    #         # compute average power of each band of interest
    #         n_chan = len(self.channels)
    #         lfp_power = np.zeros((n_chan * len(self.bands), 1))
    #         for idx, band in enumerate(self.bands):
    #             if self.extractor_kwargs['no_log']:
    #                 lfp_power[idx*n_chan : (idx+1)*n_chan, 0] = np.mean(psd_est[:, self.fft_inds[idx]], axis=1)
    #             else:
    #                 lfp_power[idx*n_chan : (idx+1)*n_chan, 0] = np.mean(np.log10(psd_est[:, self.fft_inds[idx]] + self.epsilon), axis=1)
            
    #         return lfp_power
        
    def extract_features(self, cont_samples):
        '''
        Extract spectral features from a block of time series samples

        Parameters
        ----------
        cont_samples : np.ndarray of shape (n_channels, n_samples)
            Raw voltage time series (one per channel) from which to extract spectral features 

        Returns
        -------
        lfp_power : np.ndarray of shape (n_channels * n_features, 1)
            Multi-band power estimates for each channel, for each band specified when the feature extractor was instantiated.
        '''
        assert int(self.win_len * self.fs) == cont_samples.shape[1]
        
        freqs, time, psd_est = aopy.analysis.calc_mt_tfr(cont_samples.T, *self.npk, self.fs, 1)
        if ('no_mean' in self.extractor_kwargs) and (self.extractor_kwargs['no_mean'] is True):
            return psd_est.reshape(psd_est.shape[0]*psd_est.shape[1], 1)
        else:
            # compute average power of each band of interest
            lfp_power = aopy.analysis.get_tfr_feats(freqs, psd_est, self.bands, not self.extractor_kwargs['no_log'], epsilon=self.epsilon)
            if lfp_power.ndim > 1:
                return lfp_power.reshape(lfp_power.shape[0]*lfp_power.shape[1], 1)
            else:
                return lfp_power.reshape(lfp_power.size, 1)
                
    def __call__(self, start_time, *args, **kwargs):
        '''
        Parameters
        ----------
        start_time : float 
            Absolute time from the task event loop. This is unused by LFP extractors in their current implementation
            and only passed in to ensure that function signatures are the same across extractors.
        *args, **kwargs : optional positional/keyword arguments
            These are passed to the source, or ignored (not needed for this extractor).

        Returns
        -------
        dict
            Extracted features to be saved in the task.         
        '''
        cont_samples = self.get_cont_samples(*args, **kwargs)  # dims of channels x time
        lfp_power = self.extract_features(cont_samples)

        return dict(lfp_power=lfp_power)

    @classmethod
    def extract_from_file(cls, files, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
        '''
        Compute binned spike count features

        Parameters
        ----------
        files : dict
            Data files used to train the decoder. Should contain exactly one type of neural data file (e.g., Plexon, Blackrock, TDT)
        neurows: np.ndarray of shape (T,)
            Timestamps in the plexon time reference corresponding to bin boundaries
        binlen: float
            Length of time over which to sum spikes from the specified cells
        units: np.ndarray of shape (N, 2)
            List of units that the decoder will be trained on. The first column specifies the electrode number and the second specifies the unit on the electrode
        extractor_kwargs: dict 
            Any additional parameters to be passed to the feature extractor. This function is agnostic to the actual extractor utilized
        strobe_rate: 60.0
            The rate at which the task sends the sync pulse to the plx file

        Returns
        -------
        spike_counts : np.ndarray of shape (N, T)
            Spike counts binned over the length of the datafile.
        units : 
            Not used by this type of extractor, just passed back from the input argument to make the outputs consistent with spike count extractors
        extractor_kwargs : dict
            Parameters used to instantiate the feature extractor, to be stored 
            along with the trained decoder so that the exact same feature extractor can be re-created at runtime.
        '''
    
        # interpolate between the rows to 180 Hz
        if binlen < 1./strobe_rate:
            interp_rows = []
            neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
            for r1, r2 in zip(neurows[:-1], neurows[1:]):
                interp_rows += list(np.linspace(r1, r2, 4)[1:])
            interp_rows = np.array(interp_rows)
        else:
            step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
            interp_rows = neurows[::step]
        
        # create extractor object
        f_extractor = LFPMTMPowerExtractor(None, **extractor_kwargs)
        extractor_kwargs = f_extractor.extractor_kwargs

        win_len  = f_extractor.win_len
        bands    = f_extractor.bands
        channels = f_extractor.channels
        fs       = f_extractor.fs
        print(('bands:', bands))
        print(('win_len:', win_len))

        n_itrs = len(interp_rows)
        n_chan = len(channels)
        lfp_power = np.zeros((n_itrs, n_chan * len(bands)))

        # get the samples
        if 'plexon' in files:
            from plexon import plexfile
            plx = plexfile.openFile(files['plexon'].encode('utf-8'))
            lfp = plx.lfp[:].data[:, channels-1]

        elif 'blackrock' in files:
            raise NotImplementedError

        elif 'ecube' in files:
            from riglib.ecube.file import load_lfp
            data = load_lfp(str(files['ecube']))
            lfp = data[:, [c-1 for c in channels]]
        
        # for i, t in enumerate(interp_rows):
        #     cont_samples = plx.lfp[t-win_len:t].data[:, channels-1]
        #     lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
        n_pts = int(win_len * fs)
        for i, t in enumerate(interp_rows):
            # try:
            sample_num = int(t * fs)
            cont_samples = lfp[max(0,sample_num-n_pts):min(lfp.shape[0], sample_num), :]
            if cont_samples.shape[0] < n_pts:
                # maybe don't count these ones?
                continue
            lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
            # except:
            #     print("Error with LFP decoder training")
            #     print((i, t))


        # TODO -- discard any channel(s) for which the log power in any frequency 
        #   bands was ever equal to -inf (i.e., power was equal to 0)
        # or, perhaps just add a small epsilon inside the log to avoid this
        # then, remember to do this:  extractor_kwargs['channels'] = channels
        #   and reset the units variable

        return lfp_power, units, extractor_kwargs


#########################################################
##### Reconstruction extractors, used in test cases #####
#########################################################
class ReplaySpikeCountsExtractor(BinnedSpikeCountsExtractor):
    '''
    A "feature extractor" that replays spike counts from an HDF file
    '''
    feature_type = 'spike_counts'
    def __init__(self, hdf_table, source='spike_counts', cycle_rate=60.0, units=[]):
        '''
        Parameters
        ----------
        hdf_table : HDF table
            Data table to replay. Usually the 'task' table. 
        source : string, optional, default=spike_counts
            Column of the HDF table to replay
        cycle_rate : float, optional, default=60.0
            Rate at which the task FSM "cycles", i.e., the rate at which the task will ask for new observations
        units : iterable, optional, default=[]
            Names (channel, unit) of the units. If none specified, some fake names are created
    
        Returns
        -------
        ReplaySpikeCountsExtractor instance
        '''
        self.idx = 0
        self.hdf_table = hdf_table
        self.source = source
        self.units = units
        self.n_subbins = hdf_table[0][source].shape[1]
        self.last_get_spike_counts_time = 0
        self.cycle_rate = cycle_rate

        n_units = hdf_table[0]['spike_counts'].shape[0]
        self.feature_dtype = [('spike_counts', 'u4', (n_units, self.n_subbins)), 
                              ('bin_edges', 'f8', 2)]

    def get_spike_ts(self):
        '''
        Make up fake timestamps to go with the spike counts extracted from the HDF file
        '''
        from . import sim_neurons

        # Get counts from HDF file
        counts = self.hdf_table[self.idx][self.source]
        n_subbins = counts.shape[1]

        # Convert counts to timestamps between (self.idx*1./cycle_rate, (self.idx+1)*1./cycle_rate)
        # NOTE: this code is mostly copied from riglib.bmi.sim_neurons.CLDASimPointProcessEnsemble
        ts_data = []
        cycle_rate = self.cycle_rate
        for k in range(n_subbins):
            fake_time = (self.idx - 1) * 1./cycle_rate + (k + 0.5)*1./cycle_rate*1./n_subbins
            nonzero_units, = np.nonzero(counts[:,k])
            for unit_ind in nonzero_units:
                n_spikes = counts[unit_ind, k]
                for m in range(n_spikes):
                    ts = (fake_time, self.units[unit_ind, 0], self.units[unit_ind, 1], fake_time)
                    ts_data.append(ts)

        ts_dtype_new = sim_neurons.ts_dtype_new
        return np.array(ts_data, dtype=ts_dtype_new)

    def get_bin_edges(self, ts):
        '''
        Get the first and last timestamp of spikes in the current "bin" as saved in the HDF file
        '''
        return self.hdf_table[self.idx]['bin_edges']

    def __call__(self, *args, **kwargs):
        '''
        See BinnedSpikeCountsExtractor.__call__ for documentation
        '''
        output = super(ReplaySpikeCountsExtractor, self).__call__(*args, **kwargs)
        if not np.array_equal(output['spike_counts'], self.hdf_table[self.idx][self.source]):
            print(("spike binning error: ", self.idx))
        self.idx += 1 
        return output

class ReplayLFPPowerExtractor(LFPMTMPowerExtractor):
    '''
    A "feature extractor" that replays LFP power estimates from an HDF file
    '''

    def __init__(self, hdf_table, source='lfp_power'):
        '''    
        Constructor for ReplayLFPPowerExtractor

        Parameters
        ----------
        hdf_table : HDF table
            Data table to replay. Usually the 'task' table. 
        source : string, optional, default=spike_counts
            Column of the HDF table to replay
    
        Returns
        -------
        ReplayLFPPowerExtractor instance
        '''
        self.idx = 0
        self.hdf_table = hdf_table
        self.source = source
        self.n_subbins = hdf_table[0][source].shape[1]
        self.last_get_spike_counts_time = 0

        n_units = hdf_table[0][source].shape[0]
        self.feature_dtype = [('lfp_power', 'f8', (n_units, self.n_subbins)), ]

    def __call__(self, *args, **kwargs):
        '''    
        See LFPMTMPowerExtractor.__call__ for documentation
        '''
        output = self.hdf_table[self.idx][self.source]
        self.idx += 1 
        return dict(lfp_power=output)

#################################
##### Simulation extractors #####
#################################
class SimBinnedSpikeCountsExtractor(BinnedSpikeCountsExtractor):
    '''
    Spike count features are generated by a population of synthetic neurons
    '''
    feature_type = 'spike_counts'

    def __init__(self, input_device, encoder, n_subbins, units, task=None):
        '''
        Constructor for SimBinnedSpikeCountsExtractor

        Parameters
        ----------
        input_device: object with a "calc_next_state" method
            Generate the "intended" next state, e.g., by feedback control policy
        encoder: callable with 1 argument
            Maps the "control" input into the spike timestamps of a set of neurons
        n_subbins:
            Number of subbins to divide the spike counts into, e.g. 3 are necessary for the PPF
        units: np.ndarray of shape (N, 2)
            Each row of the array corresponds to (channel, unit)

        Returns
        -------
        SimBinnedSpikeCountsExtractor instance
        '''
        self.input_device = input_device
        self.encoder = encoder
        self.n_subbins = n_subbins
        self.units = units
        self.last_get_spike_counts_time = 0
        self.feature_dtype = [('spike_counts', 'f8', (len(units), n_subbins)), ('bin_edges', 'f8', 2),
            ('ctrl_input', 'f8', self.encoder.C.shape[1])]
        self.task = task
        self.sim_ctrl = np.zeros((self.encoder.C.shape[1]))
        
    def get_spike_ts(self):
        '''
        see BinnedSpikeCountsExtractor.get_spike_ts for docs
        '''
        current_state = self.task.get_current_state()
        target_state = self.task.get_target_BMI_state()
        ctrl = self.input_device.calc_next_state(current_state, target_state)
        #print current_state.T, target_state.T, ctrl.T
        self.sim_ctrl = ctrl
        ts_data = self.encoder(ctrl)
        return ts_data

class SimPowerExtractor(LFPMTMPowerExtractor):
    '''
    This extractor pretends to generate MTMPower estimates but really it just  
    fetches the powers from a given encoder at the appropriate freq bands
    '''
    def __init__(self, input_device, encoder, channels=[], bands=default_bands, task=None, **kwargs):
        '''
        Constructor for SimPowerExtractor

        Parameters
        ----------
        input_device: object with a "calc_next_state" method
            Generate the "intended" next state, e.g., by feedback control policy
        encoder: callable with 1 argument
            Maps the "control" input into the spike timestamps of a set of neurons
        channels : list 
            LFP electrode indices to use for feature extraction
        bands : list of tuples
            Each tuple defines a frequency band of interest as (start frequency, end frequency) 

        Returns
        -------
        SimPowerExtractor instance
        '''
        self.input_device = input_device
        self.encoder = encoder
        self.channels = channels
        self.bands = bands
        self.feature_dtype = ('lfp_power', 'f8', (len(channels)*len(bands), 1))
        self.task = task
        extractor_kwargs = dict()
        extractor_kwargs['channels']   = channels
        extractor_kwargs['bands']      = bands
        self.extractor_kwargs = extractor_kwargs
        
    def get_cont_samples(self, *args, **kwargs):
        '''
        see LFPMTMPowerExtractor.get_cont_samples for docs
        '''
        current_state = self.task.get_current_state()
        target_state = self.task.get_target_BMI_state()
        ctrl = self.input_device.calc_next_state(current_state, target_state)
        cont_data = self.encoder(ctrl)
        return cont_data

    def __call__(self, start_time, *args, **kwargs):
        '''
        Parameters
        ----------
        start_time : float 
            Absolute time from the task event loop. This is unused by LFP extractors in their current implementation
            and only passed in to ensure that function signatures are the same across extractors.
        *args, **kwargs : optional positional/keyword arguments
            These are passed to the source, or ignored (not needed for this extractor).

        Returns
        -------
        dict
            Extracted features to be saved in the task.         
        '''
        lfp_power = self.get_cont_samples(*args, **kwargs)  # dims of channels x bands
        return dict(lfp_power=lfp_power)

class SimDirectObsExtractor(SimBinnedSpikeCountsExtractor):
    '''
    This extractor just passes back the observation vector generated by the encoder
    '''
    def __call__(self, start_time, *args, **kwargs):
        y_t = self.get_spike_ts(*args, **kwargs)
        return dict(spike_counts=np.reshape(y_t, (len(self.units),self.n_subbins)))


#############################################
##### Feature extractors in development #####
#############################################
class LFPButterBPFPowerExtractor(object):
    '''
    Computes log power of the LFP in different frequency bands (for each 
    channel) in time-domain using Butterworth band-pass filters.
    '''

    feature_type = 'lfp_power'

    def __init__(self, source, channels=[], bands=default_bands, win_len=0.2, filt_order=5, fs=1000):
        self.feature_dtype = ('lfp_power', 'u4', (len(channels)*len(bands), 1))

        self.source = source
        self.channels = channels
        self.bands = bands
        self.win_len = win_len  # secs
        self.filt_order = filt_order
        if source is not None:
            self.fs = source.source.update_freq
        else:
            self.fs = fs

        extractor_kwargs = dict()
        extractor_kwargs['channels']   = self.channels
        extractor_kwargs['bands']      = self.bands
        extractor_kwargs['win_len']    = self.win_len
        extractor_kwargs['filt_order'] = self.filt_order
        extractor_kwargs['fs']         = self.fs
        self.extractor_kwargs = extractor_kwargs

        self.n_pts = int(self.win_len * self.fs)
        self.filt_coeffs = dict()
        for band in bands:
            nyq = 0.5 * self.fs
            low = band[0] / nyq
            high = band[1] / nyq
            self.filt_coeffs[band] = butter(self.filt_order, [low, high], btype='band')  # returns (b, a)

        self.epsilon = 1e-9

        self.last_get_lfp_power_time = 0  # TODO -- is this variable necessary for LFP?

    def get_cont_samples(self, *args, **kwargs):
        return self.source.get(self.n_pts, self.channels)

    def extract_features(self, cont_samples):
        n_chan = len(self.channels)
        
        lfp_power = np.zeros((n_chan * len(self.bands), 1))
        for i, band in enumerate(self.bands):
            b, a = self.filt_coeffs[band]
            y = lfilter(b, a, cont_samples)
            lfp_power[i*n_chan:(i+1)*n_chan] = np.log((1. / self.n_pts) * np.sum(y**2, axis=1) + self.epsilon).reshape(-1, 1)

        return lfp_power

    def __call__(self, start_time, *args, **kwargs):
        cont_samples = self.get_cont_samples(*args, **kwargs)  # dims of channels x time
        lfp_power = self.extract_features(cont_samples)

        self.last_get_lfp_power_time = start_time
        
        return dict(lfp_power=lfp_power)

    @classmethod
    def extract_from_file(cls, files, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
        '''Compute lfp power features from a blackrock data file.'''

        nsx_fnames = [name for name in files['blackrock'] if '.ns' in name]

        # interpolate between the rows to 180 Hz
        if binlen < 1./strobe_rate:
            interp_rows = []
            neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
            for r1, r2 in zip(neurows[:-1], neurows[1:]):
                interp_rows += list(np.linspace(r1, r2, 4)[1:])
            interp_rows = np.array(interp_rows)
        else:
            step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
            interp_rows = neurows[::step]

        # TODO -- for now, use .ns3 or .ns2 file (2 kS/s)
        nsx_fname = None
        for fname in nsx_fnames:
            if '.ns3' in fname:
                nsx_fname = fname
                fs_ = 2000
        if nsx_fname is None:
            for fname in nsx_fnames:
                if '.ns2' in fname:
                    nsx_fname = fname
                    fs_ = 1000
        
        if nsx_fname is None:
            raise Exception('Need an nsx file --> .ns2 or .ns3 is acceptable. Higher nsx files yield memory errors')
        extractor_kwargs['fs'] = fs_

        # default order of 5 seems to cause problems when fs > 1000
        extractor_kwargs['filt_order'] = 3

        if nsx_fname[-4:] == '.hdf':
            nsx_hdf_fname = nsx_fname
        else:
            nsx_hdf_fname = nsx_fname + '.hdf'
            if not os.path.isfile(nsx_hdf_fname):
                # convert .nsx file to hdf file using Blackrock's n2h5 utility
                from db.tracker import models
                models.parse_blackrock_file(None, [nsx_fname], )

        import h5py
        nsx_hdf = h5py.File(nsx_hdf_fname, 'r')

        # create extractor object
        f_extractor = LFPButterBPFPowerExtractor(None, **extractor_kwargs)
        extractor_kwargs = f_extractor.extractor_kwargs

        win_len  = f_extractor.win_len
        bands    = f_extractor.bands
        channels = f_extractor.channels
        fs       = f_extractor.fs

        n_itrs = len(interp_rows)
        n_chan = len(channels)
        lfp_power = np.zeros((n_itrs, n_chan * len(bands)))
        n_pts = int(win_len * fs)

        # for i, t in enumerate(interp_rows):
        #     sample_num = int(t * fs)
        #     # cont_samples = np.zeros((n_chan, n_pts))

        #     # for j, chan in enumerate(channels):
        #     #     chan_str = str(chan).zfill(5)
        #     #     path = 'channel/channel%s/continuous_set' % chan_str
        #     #     cont_samples[j, :] = nsx_hdf.get(path).value[sample_num-n_pts:sample_num]
        #     cont_samples = abs(np.random.randn(n_chan, n_pts))

        #     feats = f_extractor.extract_features(cont_samples).T
        #     print feats
        #     lfp_power[i, :] = f_extractor.extract_features(cont_samples).T
        
        print(('*' * 40))
        print('WARNING: replacing LFP values from .ns3 file with random values!!')
        print(('*' * 40))

        lfp_power = abs(np.random.randn(n_itrs, n_chan * len(bands)))

        # TODO -- discard any channel(s) for which the log power in any frequency 
        #   bands was ever equal to -inf (i.e., power was equal to 0)
        # or, perhaps just add a small epsilon inside the log to avoid this
        # then, remember to do this:  extractor_kwargs['channels'] = channels
        #   and reset the units variable
        
        return lfp_power, units, extractor_kwargs


class AIMTMPowerExtractor(LFPMTMPowerExtractor):
    ''' Multitaper extractor for Plexon analog input channels'''

    feature_type = 'ai_power'

    def __init__(self, source, channels=[], bands=default_bands, win_len=0.2, NW=3, fs=1000, **kwargs):
        #self.feature_dtype = ('lfp_power', 'f8', (len(channels)*len(bands), 1))

        self.source = source
        self.channels = channels
        self.bands = bands
        self.win_len = win_len
        self.NW = NW
        if source is not None:
            self.fs = source.source.update_freq
        else:
            self.fs = fs

        extractor_kwargs = dict()
        extractor_kwargs['channels'] = self.channels
        extractor_kwargs['bands']    = self.bands
        extractor_kwargs['win_len']  = self.win_len
        extractor_kwargs['NW']       = self.NW
        extractor_kwargs['fs']       = self.fs

   
        extractor_kwargs['no_log']  = 'no_log' in kwargs and kwargs['no_log']==True #remove log calculation
        extractor_kwargs['no_mean'] = 'no_mean' in kwargs and kwargs['no_mean']==True #r
        self.extractor_kwargs = extractor_kwargs

        self.n_pts = int(self.win_len * self.fs)
        self.nfft = 2**int(np.ceil(np.log2(self.n_pts)))  # nextpow2(self.n_pts)
        fft_freqs = np.arange(0., fs, float(fs)/self.nfft)[:self.nfft/2 + 1]
        self.fft_inds = dict()
        for band_idx, band in enumerate(bands):
            self.fft_inds[band_idx] = [freq_idx for freq_idx, freq in enumerate(fft_freqs) if band[0] <= freq < band[1]]

        extractor_kwargs['fft_inds']       = self.fft_inds
        extractor_kwargs['fft_freqs']      = fft_freqs
        
        self.epsilon = 1e-9

        if extractor_kwargs['no_mean']: #Used in lfp 1D control task
            self.feature_dtype = ('ai_power', 'f8', (len(channels)*len(fft_freqs), 1))
        else: #Else: 
            self.feature_dtype = ('ai_power', 'f8', (len(channels)*len(bands), 1))

    def __call__(self, start_time, *args, **kwargs):
        cont_samples = self.get_cont_samples(*args, **kwargs)  # dims of channels x time
        #cont_samples = np.random.randn(len(self.channels), self.n_pts)  # change back!
        lfp_power = self.extract_features(cont_samples)

        return dict(ai_power=lfp_power)



    @classmethod
    def extract_from_file(cls, files, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
        '''
        Compute binned spike count features

        Parameters
        ----------
        plx: neural data file instance
        neurows: np.ndarray of shape (T,)
            Timestamps in the plexon time reference corresponding to bin boundaries
        binlen: float
            Length of time over which to sum spikes from the specified cells
        units: np.ndarray of shape (N, 2)
            List of units that the decoder will be trained on. The first column specifies the electrode number and the second specifies the unit on the electrode
        extractor_kwargs: dict 
            Any additional parameters to be passed to the feature extractor. This function is agnostic to the actual extractor utilized
        strobe_rate: 60.0
            The rate at which the task sends the sync pulse to the plx file

        Returns
        -------
        '''
        if 'plexon' in files:
            from plexon import plexfile
            plx = plexfile.openFile(str(files['plexon']))

            # interpolate between the rows to 180 Hz
            if binlen < 1./strobe_rate:
                interp_rows = []
                neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
                for r1, r2 in zip(neurows[:-1], neurows[1:]):
                    interp_rows += list(np.linspace(r1, r2, 4)[1:])
                interp_rows = np.array(interp_rows)
            else:
                step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
                interp_rows = neurows[::step]


            # create extractor object
            f_extractor = AIMTMPowerExtractor(None, **extractor_kwargs)
            extractor_kwargs = f_extractor.extractor_kwargs

            win_len  = f_extractor.win_len
            bands    = f_extractor.bands
            channels = f_extractor.channels
            fs       = f_extractor.fs
            print(('bands:', bands))

            n_itrs = len(interp_rows)
            n_chan = len(channels)
            lfp_power = np.zeros((n_itrs, n_chan * len(bands)))
            
            # for i, t in enumerate(interp_rows):
            #     cont_samples = plx.lfp[t-win_len:t].data[:, channels-1]
            #     lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
            lfp = plx.lfp[:].data[:, channels-1]
            n_pts = int(win_len * fs)
            for i, t in enumerate(interp_rows):
                try:
                    sample_num = int(t * fs)
                    cont_samples = lfp[sample_num-n_pts:sample_num, :]
                    lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
                except:
                    print("Error with LFP decoder training")
                    print((i, t))
                    pass


            # TODO -- discard any channel(s) for which the log power in any frequency 
            #   bands was ever equal to -inf (i.e., power was equal to 0)
            # or, perhaps just add a small epsilon inside the log to avoid this
            # then, remember to do this:  extractor_kwargs['channels'] = channels
            #   and reset the units variable

            return lfp_power, units, extractor_kwargs

        elif 'blackrock' in files:
            raise NotImplementedError

        elif 'ecube' in files:
            raise NotImplementedError


class AIAmplitudeExtractor(object):
    '''
    Computes the analog input channel amplitude. Out of date...
    '''

    feature_type = 'ai_amplitude'

    def __init__(self, source, channels=[], win_len=0.1, fs=1000):
        self.feature_dtype = ('emg_amplitude', 'u4', (len(channels), 1))

        self.source = source
        self.channels = channels
        self.win_len = win_len
        if source is not None:
            self.fs = source.source.update_freq
        else:
            self.fs = fs

        extractor_kwargs = dict()
        extractor_kwargs['channels'] = self.channels
        extractor_kwargs['fs']       = self.fs
        extractor_kwargs['win_len']  = self.win_len
        self.extractor_kwargs = extractor_kwargs

        self.n_pts = int(self.win_len * self.fs)

    def get_cont_samples(self, *args, **kwargs):
        return self.source.get(self.n_pts, self.channels)

    def extract_features(self, cont_samples):
        n_chan = len(self.channels)
        emg_amplitude = np.mean(cont_samples,axis=1)
        emg_amplitude = emg_amplitude[:,None]
        return emg_amplitude

    def __call__(self, start_time, *args, **kwargs):
        cont_samples = self.get_cont_samples(*args, **kwargs)  # dims of channels x time
        emg = self.extract_features(cont_samples)
        return emg, None

class WaveformClusterCountExtractor(FeatureExtractor):
    feature_type = 'cluster_counts'
    def __init__(self, source, gmm_model_params, n_subbins=1, units=[]):
        self.feature_dtype = [('cluster_counts', 'f8', (len(units), n_subbins)), ('bin_edges', 'f8', 2)]

        self.source = source
        self.gmm_model_params = gmm_model_params        
        self.n_subbins = n_subbins
        self.units = units
        self.n_units = len(units)

        extractor_kwargs = dict()
        extractor_kwargs['n_subbins']        = self.n_subbins
        extractor_kwargs['units']            = self.units
        extractor_kwargs['gmm_model_params'] = gmm_model_params
        self.extractor_kwargs = extractor_kwargs

        self.last_get_spike_counts_time = 0        

    def get_spike_data(self):
        '''
        Get the spike timestamps from the neural data source. This function has no type checking, 
        i.e., it is assumed that the Extractor object was created with the proper source
        '''
        return self.source.get()

    def get_bin_edges(self, ts):
        '''
        Determine the first and last spike timestamps to allow HDF files 
        created by the BMI to be semi-synchronized with the neural data file
        '''
        if len(ts) == 0:
            bin_edges = np.array([np.nan, np.nan])
        else:
            min_ind = np.argmin(ts['ts'])
            max_ind = np.argmax(ts['ts'])
            bin_edges = np.array([ts[min_ind]['ts'], ts[max_ind]['ts']])

    def __call__(self, start_time, *args, **kwargs):

        spike_data = self.get_spike_data()
        if len(spike_data) == 0:
            counts = np.zeros([len(self.units), self.n_subbins])
        elif self.n_subbins > 1:
            subbin_edges = np.linspace(self.last_get_spike_counts_time, start_time, self.n_subbins+1)

            # Decrease the first subbin index to include any spikes that were
            # delayed in getting to the task layer due to threading issues
            # An acceptable delay is 1 sec or less. Realistically, most delays should be
            # on the millisecond order
            # subbin_edges[0] -= 1
            # subbin_inds = np.digitize(spike_data['arrival_ts'], subbin_edges)
            # counts = np.vstack([bin_spikes(ts[subbin_inds == k], self.units) for k in range(1, self.n_subbins+1)]).T
            raise NotImplementedError
        else:
            # TODO pull the waveforms
            waveforms = []

            # TODO determine p(class) for each waveform against the model params
            counts = np.zeros(self.n_units)
            wf_class_probs = []
            for wf in waveforms:
                raise NotImplementedError
            
            # counts = bin_spikes(ts, self.units).reshape(-1, 1)

        counts = np.array(counts, dtype=np.uint32)
        bin_edges = self.get_bin_edges(ts)
        self.last_get_spike_counts_time = start_time

        return dict(spike_counts=counts, bin_edges=bin_edges)        

    @classmethod
    def extract_from_file(cls, files, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):        
        from sklearn.mixture import GMM
        if 'plexon' in files:
            from plexon import plexfile
            plx = plexfile.openFile(str(files['plexon']))        
        
            channels = units[:,0]
            channels = np.unique(channels)
            np.sort(channels)

            spike_chans = plx.spikes[:].data['chan']
            spike_times = plx.spikes[:].data['ts']
            waveforms = plx.spikes[:].waveforms

            # construct the feature matrix (n_timepoints, n_units)
            # interpolate between the rows to 180 Hz
            if binlen < 1./strobe_rate:
                interp_rows = []
                neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
                for r1, r2 in zip(neurows[:-1], neurows[1:]):
                    interp_rows += list(np.linspace(r1, r2, 4)[1:])
                interp_rows = np.array(interp_rows)
            else:
                step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
                interp_rows = neurows[::step]

            # digitize the spike timestamps into interp_rows
            spike_bin_ind = np.digitize(spike_times, interp_rows)
            spike_counts = np.zeros(len(interp_rows), n_units)

            for ch in channels:
                ch_waveforms = waveforms[spike_chans == ch]

                # cluster the waveforms using a GMM
                # TODO pick the number of components in an unsupervised way!
                n_components = len(np.nonzero(units[:,0] == ch)[0])
                gmm = GMM(n_components=n_components)
                gmm.fit(ch_waveforms)

                # store the cluster probabilities back in the same order that the waveforms were extracted
                wf_probs = gmm.predict_proba(ch_waveforms)

                ch_spike_bin_inds = spike_bin_ind[spike_chans == ch]
                ch_inds, = np.nonzero(units[:,0] == ch)

                # TODO don't assume the units are sorted!
                for bin_ind, wf_prob in zip(ch_spike_bin_inds, wf_probs):
                    spike_counts[bin_ind, ch_inds] += wf_prob


            # discard units that never fired at all
            unit_inds, = np.nonzero(np.sum(spike_counts, axis=0))
            units = units[unit_inds,:]
            spike_counts = spike_counts[:, unit_inds]
            extractor_kwargs['units'] = units

            return spike_counts, units, extractor_kwargs
        elif 'blackrock' in files:
            raise NotImplementedError('Not implemented for blackrock/TDT data yet!')

        elif 'ecube' in files:
            raise NotImplementedError

def get_butter_bpf_lfp_power(plx, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
    '''
    Compute lfp power features -- corresponds to LFPButterBPFPowerExtractor.


    '''
    
    # interpolate between the rows to 180 Hz
    if binlen < 1./strobe_rate:
        interp_rows = []
        neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
        for r1, r2 in zip(neurows[:-1], neurows[1:]):
            interp_rows += list(np.linspace(r1, r2, 4)[1:])
        interp_rows = np.array(interp_rows)
    else:
        step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
        interp_rows = neurows[::step]


    # create extractor object
    f_extractor = extractor.LFPButterBPFPowerExtractor(None, **extractor_kwargs)
    extractor_kwargs = f_extractor.extractor_kwargs

    win_len  = f_extractor.win_len
    bands    = f_extractor.bands
    channels = f_extractor.channels
    fs       = f_extractor.fs
        
    n_itrs = len(interp_rows)
    n_chan = len(channels)
    lfp_power = np.zeros((n_itrs, n_chan * len(bands)))
    # for i, t in enumerate(interp_rows):
    #     cont_samples = plx.lfp[t-win_len:t].data[:, channels-1]
    #     lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
    lfp = plx.lfp[:].data[:, channels-1]
    n_pts = int(win_len * fs)
    for i, t in enumerate(interp_rows):
        sample_num = int(t * fs)
        cont_samples = lfp[sample_num-n_pts:sample_num, :]
        lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
    
    # TODO -- discard any channel(s) for which the log power in any frequency 
    #   bands was ever equal to -inf (i.e., power was equal to 0)
    # or, perhaps just add a small epsilon inside the log to avoid this
    # then, remember to do this:  extractor_kwargs['channels'] = channels
    #   and reset the units variable

    return lfp_power, units, extractor_kwargs


def get_mtm_lfp_power(plx, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
    '''
    Compute lfp power features -- corresponds to LFPMTMPowerExtractor.


    '''
    
    # interpolate between the rows to 180 Hz
    if binlen < 1./strobe_rate:
        interp_rows = []
        neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
        for r1, r2 in zip(neurows[:-1], neurows[1:]):
            interp_rows += list(np.linspace(r1, r2, 4)[1:])
        interp_rows = np.array(interp_rows)
    else:
        step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
        interp_rows = neurows[::step]


    # create extractor object
    f_extractor = extractor.LFPMTMPowerExtractor(None, **extractor_kwargs)
    extractor_kwargs = f_extractor.extractor_kwargs

    win_len  = f_extractor.win_len
    bands    = f_extractor.bands
    channels = f_extractor.channels
    fs       = f_extractor.fs
    print(('bands:', bands))

    n_itrs = len(interp_rows)
    n_chan = len(channels)
    lfp_power = np.zeros((n_itrs, n_chan * len(bands)))
    
    # for i, t in enumerate(interp_rows):
    #     cont_samples = plx.lfp[t-win_len:t].data[:, channels-1]
    #     lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
    lfp = plx.lfp[:].data[:, channels-1]
    n_pts = int(win_len * fs)
    for i, t in enumerate(interp_rows):
        sample_num = int(t * fs)
        cont_samples = lfp[sample_num-n_pts:sample_num, :]
        lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T


    # TODO -- discard any channel(s) for which the log power in any frequency 
    #   bands was ever equal to -inf (i.e., power was equal to 0)
    # or, perhaps just add a small epsilon inside the log to avoid this
    # then, remember to do this:  extractor_kwargs['channels'] = channels
    #   and reset the units variable

    return lfp_power, units, extractor_kwargs

def get_emg_amplitude(plx, neurows, binlen, units, extractor_kwargs, strobe_rate=60.0):
    '''
    Compute EMG features.

    '''

    # interpolate between the rows to 180 Hz
    if binlen < 1./strobe_rate:
        interp_rows = []
        neurows = np.hstack([neurows[0] - 1./strobe_rate, neurows])
        for r1, r2 in zip(neurows[:-1], neurows[1:]):
            interp_rows += list(np.linspace(r1, r2, 4)[1:])
        interp_rows = np.array(interp_rows)
    else:
        step = int(binlen/(1./strobe_rate)) # Downsample kinematic data according to decoder bin length (assumes non-overlapping bins)
        interp_rows = neurows[::step]


    # create extractor object
    f_extractor = extractor.EMGAmplitudeExtractor(None, **extractor_kwargs)
    extractor_kwargs = f_extractor.extractor_kwargs

    win_len  = f_extractor.win_len
    channels = f_extractor.channels
    fs       = f_extractor.fs

    n_itrs = len(interp_rows)
    n_chan = len(channels)
    emg = np.zeros((n_itrs, n_chan))
    
    # for i, t in enumerate(interp_rows):
    #     cont_samples = plx.lfp[t-win_len:t].data[:, channels-1]
    #     lfp_power[i, :] = f_extractor.extract_features(cont_samples.T).T
    emgraw = plx.analog[:].data[:, channels-1]
    n_pts = int(win_len * fs)
    for i, t in enumerate(interp_rows):
        sample_num = int(t * fs)
        cont_samples = emgraw[sample_num-n_pts:sample_num, :]
        emg[i, :] = f_extractor.extract_features(cont_samples.T).T

    return emg, units, extractor_kwargs



'''
Declare which database tables are visible from Django's default admin interface. 

This file was initially created by Django
'''
from .models import Task, Feature, System, TaskEntry, Calibration, DataFile, Subject, Sequence, Generator, AutoAlignment, Decoder
from django.contrib import admin
from django.db.models.signals import post_delete
from django.dispatch.dispatcher import receiver

@receiver(post_delete, sender=DataFile)
def _mymodel_delete(sender, instance, **kwargs):
    instance.remove()
    
admin.site.register(Task)
admin.site.register(Feature)
admin.site.register(System)
admin.site.register(TaskEntry)
admin.site.register(Calibration)
admin.site.register(DataFile)
admin.site.register(Subject)
admin.site.register(Sequence)
admin.site.register(Generator)
admin.site.register(AutoAlignment)
admin.site.register(Decoder)
